 /***************************************************************************
 *
 * Copyright (C) 2007-2009  SMSC
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ***************************************************************************
 * File: smsc9420_eth.c
 ***************************************************************************
 * History:
 *
 * 09/18/2007, Nancy Lin, Version 0.01
 *  This is the first version of LAN9420 PCI driver. 
 *
 * 09/20/2007, Nancy Lin,  Version 1.00
 *  Added tx_Csum, rx_Csum and scatter_gather as modifiable Module Parameters
 *      Fixed bug of RX VLAN packets
 *
 * 09/21/2007, Nancy Lin, Version 1.01
 *      Minor changes. Cleaned some unused codes.
 * 
 * 09/26/2007, Nancy Lin, Version 1.02
 *  Changed default tx_ring_size and rx_ring_size to 32.
 *
 * 10/08/2007, Nancy Lin, Version 1.03
 *  Fixed Unserviced DMAC Interrupt bug
 *  
 * 09/09/2008, A.H., Version 2.00
 *  Cleanup the driver to comply with the spec.
 *
 * 22/09/2008, Vlad Lyalikov Version 2.01
 *  Reordered SET_NETDEV_DEV and register_netdev()
 *
 * 10/10/2008, Vlad Lyalikov Version 3.00
 *  Changed almost everything...
 *  
 * 17/10/2008, Vlad Lyalikov Version 3.01
 * Added support for hw endiannes. removed MWI 
 *
 * 20/10/2008, Vlad Lyalikov Version 3.02
 * Fixed incorrect handling of bad packets
 * 
 * 21/10/2008, Vlad Lyalikov Version 3.03
 * Changed indexing on TX path DMA
 * Decreased tx_ring_size to 32.
 *
 * 22/10/2008, Vlad Lyalikov Version 3.04
 * Removed calls to queue wake or stop from
 * outside the data path. 
 * Added extensive comments on ring size parameters
 * and general driver architecture.
 *
 * 24/10/2008, Vlad Lyalikov Version 3.05
 * hard_start_xmit() now returns skb back to the queue if
 * its  unable to xmit.
 *
 * 06/11/2008, Vlad Lyalikov Version 3.06
 * Rewrote suspend/resume. 
 *
 * 07/11/2008, Vlad Lyalikov Version 3.07
 * Implemented WAKE_PHY/WAKE_MAC for 
 * all combinations of link_up/link_down
 *
 * 11/11/2008, Vlad Lyalikov Version 3.08
 * Phy sleep mode issues, preventing MAC access 
 * added isr_try_cnt to adjust number of passes in
 * DMAC portion of ISR.
 *
 * 01/20/2009 Vlad Lyalikov Version 3.09
 * Added support for 2.6.27
 * removed cmd app, redundant platform support
 * 
 */



#define DRIVER_VERSION (0x00000309UL)
#define DRV_VERSION    "3.09"



#ifndef __KERNEL__
#   define __KERNEL__
#endif


#include <linux/version.h>

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13))
#include <linux/config.h>
#endif
#include <linux/module.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#include <linux/moduleparam.h>     
#endif
#include <linux/stat.h>             
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/delay.h>
#include <linux/apm_bios.h>
#include <linux/mii.h>
#include <linux/timer.h>
#include <linux/rtnetlink.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <linux/pci.h>
#include <linux/skbuff.h>
#include <asm/unaligned.h>
#include <linux/types.h>
#include <linux/if_vlan.h>
#include "smsc_debug.h"

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0))
#error Must use linux 2.6 or newer
//The NAPI support in this driver requires linux 2.6 or newer.
typedef void irqreturn_t;
#define IRQ_NONE
#define IRQ_HANDLED
#define IRQ_RETVAL(x)
#else
#define LINUX_2_6_OR_NEWER
#endif



#define MIN_RX_RING_SIZE        (2UL)
#define MAX_RX_RING_SIZE        (4096UL)
//this is the fraction of rx_ring_size which napi_weight takes
#define DEFAULT_RX_RING_TO_NAPI_WEIGHT_FACTOR   (2UL)
#define MIN_TX_RING_SIZE        (2UL)
#define MAX_TX_RING_SIZE        (64UL)

//for explanation of this values see
//discussion above module parameters section
#define DEFAULT_TX_RING_SIZE    (32UL)
#define DEFAULT_RX_RING_SIZE    (128UL)
#define DEFAULT_INT_DEAS_10us   (50UL)

#define DEFAULT_ISR_TRY_CNT     (2UL)
#define MAX_ISR_TRY_CNT         (20UL)
//NB: because GPIO_CFG register is not thread safe,
//access to GPIO pins for debugging is subject to races between
//ISR and "tasklets".
//this should be kept in mind while using LAN9420 GPIOs in preemtpive envirnoment
#define SET_PINx(pd, x) (Lan_SetRegDW(GPIO_CFG, (Lan_GetRegDW(GPIO_CFG)) | (1 << (x))))
#define CLR_PINx(pd, x) (Lan_SetRegDW(GPIO_CFG, (Lan_GetRegDW(GPIO_CFG)) & ~(1 << (x))))
#define SET_PIN0(pd)    (SET_PINx((pd), 0))
#define CLR_PIN0(pd)    (CLR_PINx((pd), 0))
#define SET_PIN1(pd)    (SET_PINx((pd), 1))
#define CLR_PIN1(pd)    (CLR_PINx((pd), 1))
#define SET_PIN2(pd)    (SET_PINx((pd), 2))
#define CLR_PIN2(pd)    (CLR_PINx((pd), 2))


#define DRV_NAME            "lan9420"
#define DRV_DESCRIPTION     "SMSC LAN9420 Driver"
#define DRV_COPYRIGHT       "Copyright(c) 2008 SMSC"

#define BAR0                0
#define BAR1                1
#define BAR2                2
#define BAR3                3
#define BAR4                4
#define BAR5                5



#ifdef __LITTLE_ENDIAN
#define LAN9420_CPSR_ENDIAN_OFFSET      (0UL)
#else   //LE
#ifdef __BIG_ENDIAN
#define LAN9420_CPSR_ENDIAN_OFFSET      (0x200UL)
#else  //BE
#error "Eniannes of the platform is not defined"
#endif //BE
#endif  //LE

#define PCI_VENDOR_ID_9420ESS           0x1055
#define PCI_DEVICE_ID_9420ESS           0xE420

#define LAN_REGISTER_EXTENT     (0x400UL)        //CSR size in Non_prefetchable area
                                                //1Kb: low 512b-> LE, high 512-> BE; registers over 0x100-> reserved

#define HIBYTE(word)  ((u8)(((u16)(word))>>8))
#define LOBYTE(word)  ((u16)(((u16)(word))&0x00FFU))
#define HIWORD(dWord) ((u16)(((u32)(dWord))>>16))
#define LOWORD(dWord) ((u16)(((u32)(dWord))&0x0000FFFFUL))

#define Lan_GetRegDW(dwOffset)  \
    (*((volatile u32*)(pd->LanBase+dwOffset)))
#define Lan_SetRegDW(dwOffset,dwValue)  \
    ((*((volatile u32*)(pd->LanBase+dwOffset)))=(u32)(dwValue))



#define PHY_ACCESS_DELAY_TICK_us   (10UL)
#define PHY_ACCESS_MAX_WAIT_us     (50UL)


#define SMSC9420_DEFAULT_WOL_OPTS       (WAKE_MAGIC|WAKE_PHY)

#define LINK_OFF                (0x00UL)
#define LINK_SPEED_10HD         (0x01UL)
#define LINK_SPEED_10FD         (0x02UL)
#define LINK_SPEED_100HD        (0x04UL)
#define LINK_SPEED_100FD        (0x08UL)
#define LINK_SYMMETRIC_PAUSE    (0x10UL)
#define LINK_ASYMMETRIC_PAUSE   (0x20UL)
#define LINK_AUTO_NEGOTIATE     (0x40UL)

#define PKT_BUF_SZ              (VLAN_ETH_FRAME_LEN + NET_IP_ALIGN + 4)


#define MAX_MAC_DMAC_SRST_TIMEOUT_us    (2)

#define PD_ACC_ENTRY(acc)        
#define PD_ACC_EXIT(acc)
#define ACCOUNTING_ENTRY(pd,stat,type,func,line)        \
{       \
    if(pd)\
    {   \
        acc_tocken_t acc = {(pd), (stat), (type), (func),(line)};\
        PD_ACC_ENTRY(&acc); \
    }\
}

#define ACCOUNTING_EXIT(pd,stat,type,func,line)         \
{       \
    if(pd)\
    {   \
        acc_tocken_t acc = {(pd),(stat),(type),(func),(line)};\
        PD_ACC_EXIT(&acc);\
    }\
}


#define NINDEX_FLT                      (~0UL)
#define SET_NAME_INDEX(x, val)          ((x) = (val))
#define GET_NAME_INDEX(x)               (x)
#define IS_NAME_INDEX_SET(x)            (!(NINDEX_FLT == GET_NAME_INDEX((x))))
#define GET_CHIP_NAME(x)                (SMSC_CHIP_NAMES[GET_NAME_INDEX((x))])

#define GET_PD_DEBUG_MODE(x)            ((x)->debug_mode)
#define SET_PD_DEBUG_MODE(x, val)       ((x)->debug_mode = (val))
#define IS_PD_DEBUG_MODE_SET(x)         (!!GET_PD_DEBUG_MODE((x)))

#define SET_PD_REVID(pd,val)            ((pd)->dwIdRev=(val))
#define GET_PD_REVID(pd)                ((pd)->dwIdRev)
#define IS_PD_REVID_SET(pd)             (!!(GET_PD_REVID((pd))))

#define SET_PD_FPGA_REV(pd,val)         ((pd)->dwFpgaRev=(val))
#define GET_PD_FPGA_REV(pd)             ((pd)->dwFpgaRev)
#define IS_PD_FPGA_REV_SET(pd)          (!!(GET_PD_FPGA_REV((pd))))

#define SET_PD_ISR_CNT(pd, val)         ((pd)->isr_try_cnt=(val))
#define GET_PD_ISR_CNT(pd)              ((pd)->isr_try_cnt)

#define IS_CLR_TIMER_ON_RX(pd)          ((pd)->clr_deas_timer_on_rx)
#define SET_CLR_TIMER_ON_RX(pd)         ((pd)->clr_deas_timer_on_rx = 1)
#define RST_CLR_TIMER_ON_RX(pd)         ((pd)->clr_deas_timer_on_rx = 0)


#define SET_PD_WOLOPTS(p, val)          ((p)->WolWakeupOpts = (val))
#define GET_PD_WOLOPTS(p)               ((p)->WolWakeupOpts)

#define SET_PD_NEEDSWAKE(p)             ((p)->NeedsWake = 1)
#define CLR_PD_NEEDSWAKE(p)             ((p)->NeedsWake = 0)
#define IS_PD_NEEDSWAKE_SET(p)              ((p)->NeedsWake)

#define SET_PD_LANBASE(p, val)          ((p)->LanBase = (val))
#define GET_PD_LANBASE(p)               ((p)->LanBase)
#define IS_PD_LANBASE_SET(p)            (!!(GET_PD_LANBASE((p))))

#define SET_PD_INTDEAS(p, val)          ((p)->IntDeas_10us = (val))
#define GET_PD_INTDEAS(p)               ((p)->IntDeas_10us)
#define IS_PD_INTDEAS_SET(p)            (!!(GET_PD_INTDEAS((p))))

#define SET_PD_DEV(p,val)               ((p)->dev = (struct net_device*)(val))
#define GET_PD_DEV(p)                   ((struct net_device*)(p)->dev)
#define IS_PD_DEV_SET(p)                (!!(GET_PD_DEV((p))))

#define SET_PD_PDEV(p,val)              ((p)->pdev = (struct pci_dev *)(val))
#define GET_PD_PDEV(p)                  ((struct pci_dev *)((p)->pdev))
#define IS_PD_PDEV_SET(p)               (!!(GET_PD_PDEV((p))))


#define SET_PD_RxCSUM(pd,val)           ((pd)->UseRxCsumHW=(val))
#define GET_PD_RxCSUM(pd)               ((pd)->UseRxCsumHW)
#define IS_PD_RxCSUM_SET(pd)            (!!(GET_PD_RxCSUM((pd))))


#define RX_FILTER_PROMISCUOUS   (1)
#define RX_FILTER_ALL_MULTICAST (2)

#define PD_GET_TXDMA_ADDR(p)                    ((p)->tx_dma_addr)


#define SMSC9420_PRIVATE_DATA_SIGNATURE (0x348BC926UL)

#define SET_PD_SIGNATURE(p)             ((p)->dwSignature = SMSC9420_PRIVATE_DATA_SIGNATURE)
#define IS_PD_SIGNATURE_SET(p)          (SMSC9420_PRIVATE_DATA_SIGNATURE == (p)->dwSignature)


#define TX_GET_HEAD(pd)         ((pd)->TxHeadIndex)
#define TX_SET_HEAD(pd, val)    ((pd)->TxHeadIndex = (val))
#define TX_GET_TAIL(pd)         ((pd)->TxTailIndex)
#define TX_SET_TAIL(pd, val)    ((pd)->TxTailIndex = (val))

#define TX_NEW_HEAD(pd)         ((TX_GET_HEAD((pd)) + 1) & ((pd)->tx_ring_size - 1))
#define TX_NEW_TAIL(pd)         ((TX_GET_TAIL((pd)) + 1) & ((pd)->tx_ring_size - 1))
#define TX_SET_FULL(pd)         ((pd)->TxFull = 1)
#define TX_IS_FULL(pd)          ((pd)->TxFull)
#define TX_CLR_FULL(pd)         ((pd)->TxFull = 0)

#define TX_INC_HEAD(pd)         \
{       \
    TX_SET_HEAD((pd), TX_NEW_HEAD((pd)));            \
    if(TX_GET_HEAD((pd)) == TX_GET_TAIL((pd)))   \
            TX_SET_FULL((pd));                    \
}
#define TX_INC_TAIL(pd)         \
{                               \
    TX_SET_TAIL((pd), TX_NEW_TAIL((pd)));       \
    TX_CLR_FULL(pd);                            \
}
#define TX_IS_EMPTY(pd)         ((TX_GET_TAIL((pd)) == TX_GET_HEAD((pd))) && \
                                   !( TX_IS_FULL((pd))))




#define GET_PD_RX_WLIMIT(p)        ((p)->RxWorkLimit)
#define SET_PD_RX_WLIMIT(p,val)    ((p)->RxWorkLimit = (ulong)(val))
#define IS_PD_RX_WLIMIT_SET(p)     (!!(GET_PD_RX_WLIMIT((p))))     
#define DEC_PD_RX_WLIMIT(p)        ( {                           \
                    if(IS_PD_RX_WLIMIT_SET((p)))\
                        {       \
                        SET_PD_RX_WLIMIT((p), GET_PD_RX_WLIMIT((p)) - 1);       \
                        }                       \
        }    )

    


/*********************************************
 DMA Controller Control and Status Registers 
**********************************************/
#define BUS_MODE            0x00UL  // Bus Mode Register 
#       define BUS_MODE_SWR_                    0x00000001UL //SC, Software Reset
#       define BUS_MODE_BLE_                   (1 << 7) //Buffer byte ordering: 0 for LE, 1 for BE
#       define BUS_MODE_DMA_BURST_LENGTH_1     (1 << 8)
#       define BUS_MODE_DMA_BURST_LENGTH_2     (1 << 9)
#       define BUS_MODE_DMA_BURST_LENGTH_4     (1 << 10)
#       define BUS_MODE_DMA_BURST_LENGTH_8     (1 << 11)
#       define BUS_MODE_DMA_BURST_LENGTH_16    (1 << 12)
#       define BUS_MODE_DMA_BURST_LENGTH_32    (1 << 13)
#       define BUS_MODE_DBO_                   (1 << 20)  //Descriptor Byte ordering: 0 for LE, 1 for BE

#define TX_POLL_DEMAND      0x04UL  // Transmit Poll Demand Register 
#define RX_POLL_DEMAND      0x08UL  // Receive Poll Demand Register 
#define RX_BASE_ADDR        0x0CUL  // Receive List Base Address Register 
#define TX_BASE_ADDR        0x10UL  // Transmit List Base Address Register 
#define DMAC_STATUS         0x14UL  // DMA Controller Status Register 
#   define DMAC_STS_TS_         0x00700000UL // Mask for the Transmit Process State
#       define DMAC_STS_TS_STOP_    0x00000000UL // Reset or Stop issued 
#       define DMAC_STS_TS_FETC_    0x00100000UL // Fetching the TX Descriptor 
#   define DMAC_STS_RS_         0x000E0000UL // Mask for the Receive Process State
#       define DMAC_STS_RS_SUSP_    0x000C0000UL // RX in Suspend mode, buffer unavailable
#   define DMAC_STS_NIS_        0x00010000UL // Normal Interrupt Summary 
#   define DMAC_STS_AIS_        0x00008000UL // Abnormal Interrupt Summary 
#   define DMAC_STS_RWT_        0x00000200UL // Receive Watchdog Timeout 
#   define DMAC_STS_RXPS_       0x00000100UL // Receive Process Stopped 
#   define DMAC_STS_RXBU_       0x00000080UL // Receive Buffer Unavailable 
#   define DMAC_STS_RX_         0x00000040UL // Receive Interrupt 
#   define DMAC_STS_TXUNF_      (1 << 5)     //Tx underflow
#   define DMAC_STS_TXBU_       0x00000004UL // Transmit Buffer Unavailable 
#   define DMAC_STS_TXPS_       0x00000002UL // Transmit Process Stopped 
#   define DMAC_STS_TX_         0x00000001UL // Transmit Interrupt 
#define DMAC_CONTROL        0x18UL  // DMA Controller Control (Operation mode) Register 
#   define DMAC_CONTROL_TTM_    (1 << 22)     //tx treshold mode 
#   define DMAC_CONTROL_SF_     (1 << 21)     //store and forward, use it!! 

#   define DMAC_CONTROL_MBO_    0x00200000UL // Must be one for normal operation
#   define DMAC_CONTROL_ST_     0x00002000UL // Start(1)/Stop(0) Transmist 
#   define DMAC_CONTROL_SR_     0x00000002UL // Start(1)/Stop(0) Receive
#   define DMAC_CONTROL_OSF_    0x00000004UL   //Operate on Second Frame
#define DMAC_INTR_ENA       0x1CUL  // DMA Controller Interrupt Enable Register 
#   define DMAC_INTR_ENA_NIS_   0x00010000UL // Enables Normal Interrupt Summary 
#   define DMAC_INTR_ENA_AIS_   0x00008000UL // Enable Abnormal Interrupt Summary 
#   define DMAC_INTR_ENA_RWT_   0x00000200UL
#   define DMAC_INTR_ENA_RXPS_  0x00000100UL
#   define DMAC_INTR_ENA_RXBU_  0x00000080UL
#   define DMAC_INTR_ENA_RX_    0x00000040UL
#   define DMAC_INTR_ENA_TXBU_  0x00000004UL
#   define DMAC_INTR_ENA_TXPS_  0x00000002UL
#   define DMAC_INTR_ENA_TX_    0x00000001UL
#define MISS_FRAME_CNTR     0x20UL  // Missed Frame Counter (RX Only) 
#define TX_BUFF_ADDR        0x50UL  // Current Host Transmit Buffer Address 
#define RX_BUFF_ADDR        0x54UL  // Current Host Receive Buffer Address 

// Transmit Descriptor Bit Defs 
#define TDES0_OWN_                  0x80000000UL // Own Bit 
#define TDES0_ERROR_SUMMARY_        0x00008000UL // Error Summary
#define TDES0_LOSS_OF_CARRIER_      0x00000800UL
#define TDES0_NO_CARRIER_           0x00000400UL
#define TDES0_LATE_COLLISION_       0x00000200UL
#define TDES0_EXCESSIVE_COLLISIONS_ 0x00000100UL
#define TDES0_HEARTBEAT_FAIL_       0x00000080UL
#define TDES0_COLLISION_COUNT_MASK_ 0x00000078UL
#define TDES0_COLLISION_COUNT_SHFT_ 3UL
#define TDES0_EXCESSIVE_DEFERRAL_   0x00000004UL
#define TDES0_DEFERRED_             0x00000001UL

#define TDES1_IC_                   0x80000000UL // Interrupt on Completion 
#define TDES1_LS_                   0x40000000UL // Last Segment 
#define TDES1_FS_                   0x20000000UL // First Segment 
#define TDES1_TXCSEN_               0x08000000UL // Checksum Enable.
#define TDES1_TER_                  (1 << 25)// Transmit End of Ring 
#define TDES1_TCH_                  0x01000000UL // Second Address Chaining

// Receive Descriptor Bit Defs 
#define RDES0_OWN_                  (0x80000000UL)  // Own Bit 
#define RDES0_FRAME_LENGTH_MASK_    (0x07FF0000UL)  // Length Mask 
#define RDES0_FRAME_LENGTH_SHFT_    (16UL)          // Used for shifting
#define RDES0_ERROR_SUMMARY_        (0x00008000UL)
#define RDES0_DESCRIPTOR_ERROR_     (0x00004000UL)
#define RDES0_LENGTH_ERROR_         (0x00001000UL)
#define RDES0_RUNT_FRAME_           (0x00000800UL)
#define RDES0_MULTICAST_FRAME_      (0x00000400UL)
#define RDES0_FIRST_DESCRIPTOR_     (0x00000200UL)
#define RDES0_LAST_DESCRIPTOR_      (0x00000100UL)
#define RDES0_FRAME_TOO_LONG_       (0x00000080UL)
#define RDES0_COLLISION_SEEN_       (0x00000040UL)
#define RDES0_FRAME_TYPE_           (0x00000020UL)
#define RDES0_WATCHDOG_TIMEOUT_     (0x00000010UL)
#define RDES0_MII_ERROR_            (0x00000008UL)
#define RDES0_DRIBBLING_BIT_        (0x00000004UL)
#define RDES0_CRC_ERROR_            (0x00000002UL)

#define RDES1_RER_                  0x02000000UL // Receive End of Ring 

/*********************************************
 MAC Control and Status Registers 
**********************************************/
#define MAC_CR          0x80UL // MAC Control Register 
#   define MAC_CR_RXALL_        0x80000000UL // Receive All  (this bit would deliver any
                                             //any frame to DMAC, but if the filter
                                             //fails, FF (filter fail) will be 
                                             //set in descriptor
#   define MAC_CR_DIS_RXOWN_    0x00800000UL // 1=Disable reception of frames with TXEN=1 
#   define MAC_CR_LOOPBK_       0x00200000UL // 1=Loopback mode 
#   define MAC_CR_FDPX_         0x00100000UL // 1=Full Duplex mode 
#   define MAC_CR_MCPAS_        0x00080000UL // 1=Passes all multicast packets 
#   define MAC_CR_PRMS_         0x00040000UL // 1=Passes al the frames 
#   define MAC_CR_INVFILT_      0x00020000UL // 1=inverse filtering mode 
#   define MAC_CR_PASSBAD_      0x00010000UL // 1=Pass bad frames 
#   define MAC_CR_HFILT_        0x00008000UL // 1=imperfect address filtering mode 
#   define MAC_CR_HPFILT_       0x00002000UL // 1=imperfect/hash filtering
#   define MAC_CR_LCOLL_        0x00001000UL // 1=enabled re-tx of collided frame 
#   define MAC_CR_DIS_BCAST_    0x00000800UL // 1=disables rx of broadcast 
#   define MAC_CR_DIS_RTRY_     0x00000400UL // 1=disables retry 
#   define MAC_CR_PADSTR_       0x00000100UL // 1=MAC strips the pad fiels of RX frames 
#   define MAC_CR_BOLMT_MSK     0x000000C0UL // Mask for the Backoff Limit 

#   define MAC_CR_MFCHK_        0x00000020UL // Deferral check 
#   define MAC_CR_TXEN_         0x00000008UL // 1=TX Enabled 
#   define MAC_CR_RXEN_         0x00000004UL // 1=RX Enabled 
#define ADDRH           0x84UL // MAC Address High Register 
#define ADDRL           0x88UL // MAC Address Low Register 
#define HASHH           0x8CUL // Multicast Hash Table High Register 
#define HASHL           0x90UL // Multicast Hast Table Low Register 
#define MII_ACCESS      0x94UL // MII Address Register 
#   define MII_ACCESS_MII_BUSY_         0x00000001UL    // MII Busy bit, 1=MII Busy 
#   define MII_ACCESS_MII_WRITE_        0x00000002UL    // 1 indicates a write operation 
#   define MII_ACCESS_MII_READ_         0x00000000UL    // ~of MII Write bit 
#   define MII_ACCESS_INDX_MSK_         0x000007C0UL    // Mask for the Index bits 
#   define MII_ACCESS_PHYADDR_MSK_      0x0000F8C0UL    // Mask for the PHY Address bits 
#   define MII_ACCESS_INDX_SHFT_CNT     6UL             // Used for shift operations 
#   define MII_ACCESS_PHYADDR_SHFT_CNT  11UL            // Used for shift operations 
#define MII_DATA        0x98UL // MII Data Register 
#define FLOW            0x9CUL // Flow Control Register 
#define VLAN1           0xA0UL // VLAN1 Tag Register 
#define VLAN2           0xA4UL // VLAN2 Tag Register 
#define WUFF            0xA8UL // Wake-up Frame Filter 
#define WUCSR           0xACUL // Wake-up Status and Control
    #define WUCSR_GUE_      0x0200UL //Global Unicast Enable
    #define WUCSR_WUFR_     0x0040UL //Remote Wake-Up Frame Received
    #define WUCSR_MPR_      0x0020UL //Magic Packet Received
    #define WUCSR_WAKE_EN_  0x0004UL //Wake-Up Frame Enabled
    #define WUCSR_MPEN_     0x0002UL //Magic Packet Enable

#define COE_CR          0xB0UL
    #define TX_COE_EN       0x00010000UL
    #define RX_COE_MODE     0x00000002UL
    #define RX_COE_EN       0x00000001UL

/*********************************************
 System Control and Status Registers
**********************************************/
#define ID_REV           0xC0UL /* Chip ID and Rev Register */
#define INT_CTL          0xC4UL /* Interrupt Control Register */
#   define INT_CTL_SW_INT_EN_       0x00008000UL /* Software Interrupt Enable */
#   define INT_CTL_SBERR_INT_EN_    (1 << 12)/* Bus error interrupt enable */
#   define INT_CTL_MBERR_INT_EN_    (1 << 13)/* Bus error interrupt enable */
#   define INT_CTL_GPT_INT_EN_      0X00000008UL /* GP Timer interrupt enable */
#   define INT_CTL_PHY_INT_EN_      0X00000004UL /* Phy interrupt enable */
#   define INT_CTL_PMT_INT_EN_      0X00000002UL /* Power Management interrupt enable*/

#define INT_STAT        0xC8UL /* Interrupt Status Register */
#   define INT_STAT_SW_INT_         (1 << 15)    /* Software Interrupt */
#   define INT_STAT_MBERR_INT_      (1 << 13)    /* Master PCI bus error */
#   define INT_STAT_SBERR_INT_      (1 << 12)    /* SlaveBus Error Interrupt */
#   define INT_STAT_GPIO_INTS_      ((1 << 4) | (1 << 5) | (1 << 6))    /* 3 gpio interrupts */
#   define INT_STAT_GPT_INT_        (1 << 3)     /* GP Timer Interrupt */
#   define INT_STAT_PHY_INT_        0x00000004UL /* PHY Interrupt */

#   define INT_STAT_PMT_INT_        0x00000002UL /* Power Management Interrupt */
#   define INT_STAT_DMAC_INT_       0x00000001UL /* DMAC Interrupt */

#define INT_CFG          0xCCUL /* Interrupt Configuration Register */
#   define INT_CFG_IRQ_INT_         0x00080000UL /* Master Interrupt - RO */
#   define INT_CFG_IRQ_EN_          0x00040000UL /* IRQ Enable */
#   define INT_CFG_INT_DEAS_CLR_    0x00000200UL /* Interrupt Deassertion Interval Clear */
#   define INT_CFG_INT_DEAS_MASK    0x000000FFUL /* Interrupt Deassertion time in multiples of 10us */

#define SMSC_INT_CFG_RESERVED_BITS (0xFFF3FC00UL)

#define GPIO_CFG         0xD0UL /* General Purpose IO Configuration */
    #define GPIO_CFG_LED_3_         0x40000000UL
    #define GPIO_CFG_LED_2_         0x20000000UL
    #define GPIO_CFG_LED_1_         0x10000000UL
    
    #define GPIO_CFG_GPDIR2_        0x00000400  /* GPIO2 as an output */
    #define GPIO_CFG_GPDIR1_        0x00000200  /* GPIO1 as an output */
    #define GPIO_CFG_GPDIR0_        0x00000100  /* GPIO0 as an output */

    #define GPIO_CFG_GPBUF2_        0x00040000  /* config GPIO2 output buffer as push/pull */
    #define GPIO_CFG_GPBUF1_        0x00020000  /* config GPIO1 output buffer as push/pull */
    #define GPIO_CFG_GPBUF0_        0x00010000  /* config GPIO0 output buffer as push/pull */


#define GPT_CFG          0xD4UL /* General Purpose Timer Configuration */
    #define GPT_CFG_TIMER_EN_       0x20000000UL /* Timer Enable when set */
#define GPT_CNT          0xD8UL /* General Purpose Timer current count */
#define BUS_CFG          0xDCUL /* System Bus Configuration Register */
#define PMT_CTRL         0xE0UL /* Power Management Control Register */
    #define PMT_CTRL_WOL_EN_        0x00000200UL /* Wake on Lan Enable */
    #define PMT_CTRL_ED_EN_         0x00000100UL /* Energy Detect Wake up Enable */

    #define PMT_CTRL_WUPS_MASK_     (3 << 3)
    #define PMT_CTRL_WUPS_NONE_      0
    #define PMT_CTRL_WUPS_PHY_      (1 << 3)
    #define PMT_CTRL_WUPS_MAC_      (1 << 4)
    #define PMT_CTRL_PHY_RST_       (1 << 10)

#define FPGA_REV         0xE4UL /* FPGA Revision */
#define RESERVED         0xE8UL /* PIOB Bus Error Register */
#define DEBUG1_REG       0xECUL /* Bus Master Bridge Snapshot Register */
#define TEST_REG         0xF0UL /* Test Mode Register */
#define FREE_RUN         0xF4UL /* Free run counter */




#define CSR_RXTXWEIGHT_1_1      (0)
#define CSR_RXTXWEIGHT_2_1      (1 << 25)
#define CSR_RXTXWEIGHT_3_1      (1 << 26)
#define CSR_RXTXWEIGHT_4_1      (3 << 25)


    
/*********************************************
 PHY Related Registers

**********************************************/

//#define LAN9420_PHY_ID    0x0007

#define PHY_BCR         ((u32)0U)
#   define PHY_BCR_RESET_           0x8000 /* PHY Reset */
#   define PHY_BCR_LPBK_            0x4000 /* 1 = Loopback operation */
#   define PHY_BCR_SPEED_SELECT_    0x2000 /* Speed = 100Mbps */
#   define PHY_BCR_ANEG_EN_         0x1000 /* Autonegotiation enable */
#   define PHY_BCR_PWR_DWN_         0x0800 /* 1 = General Power Down Mode */
#   define PHY_BCR_RST_ANEG_        0x0200 /* Restart Autonegotiation */
#   define PHY_BCR_DUPLEX_MODE_     0x0100 /* Full Duplex */
#   define PHY_BCR_COLL_TST_        0x0080 /* 1 = enable collision test */
#define PHY_BSR         ((u32)1U)
#   define PHY_BSR_100FDX_          0x4000 /* TX With 100 Full Duplex ability */
#   define PHY_BSR_100HDX_          0x2000 /* TX With 100 Half Duplex ability */
#   define PHY_BSR_10FDX_           0x1000 /* TX With 10 Full Duplex ability */
#   define PHY_BSR_10HDX_           0x0800 /* TX With 10 Half Duplex ability */
#   define PHY_BSR_ANEG_COMP_       0x0020 /* Auto-Negotiation complete */
#   define PHY_BSR_RMT_FLT_         0x0010 /* Remote Fault */
#   define PHY_BSR_LINK_STATUS_     0x0004 /* Link Status */
#define PHY_ID1         ((u32)2U)
#define PHY_ID2         ((u32)3U)
#define PHY_ANEG_ADV    ((u32)4U)
#   define PHY_ANEG_ADV_PAUSE_  ((u16)0x0C00)
#   define PHY_ANEG_ADV_ASYMP_  ((u16)0x0800)
#   define PHY_ANEG_ADV_SYMP_   ((u16)0x0400)
#   define PHY_ANEG_ADV_10H_    ((u16)0x20)
#   define PHY_ANEG_ADV_10F_    ((u16)0x40)
#   define PHY_ANEG_ADV_100H_   ((u16)0x80)
#   define PHY_ANEG_ADV_100F_   ((u16)0x100)
#   define PHY_ANEG_ADV_SPEED_  ((u16)0x1E0)
#define PHY_ANEG_LPA    ((u32)5U)
#   define PHY_ANEG_LPA_ASYMP_  ((u16)0x0800)
#   define PHY_ANEG_LPA_SYMP_   ((u16)0x0400)
#   define PHY_ANEG_LPA_100FDX_ ((u16)0x0100)
#   define PHY_ANEG_LPA_100HDX_ ((u16)0x0080)
#   define PHY_ANEG_LPA_10FDX_  ((u16)0x0040)
#   define PHY_ANEG_LPA_10HDX_  ((u16)0x0020)
#define PHY_ANEG_ER     (6UL)
#define PHY_REV         16
#define PHY_MCSR        17
    #define PHY_MCSR_EDPWRDOWN_     (1 << 13)
#define PHY_SM          18

#define PHY_TSTCNTL     20

#define PHY_TSTREAD1    21

#define PHY_TSTREAD2    22
#define PHY_TSTWRITE    23
#define PHY_CSIR        27
    #define PHY_SCS_AUTOMDIX_EN_        (1 << 14)
    #define PHY_SCS_AUTOMDIX_STATE_     (1 << 13)
#define PHY_SITC        28
#define PHY_INT_SRC     29

#define PHY_INT_MASK    30
    #define PHY_INT_MASK_ENERGYON_      (1 << 7)/* Energy On */
    #define PHY_INT_MASK_ANEG_COMPL_    (1 << 6)
    #define PHY_INT_MASK_REMFLT_        (1 << 5)
    #define PHY_INT_MASK_LINK_DOWN_     (1 << 4)
    #define PHY_INT_MASK_ANEG_LP_       (1 << 3)
    #define PHY_INT_MASK_PARDETFLT_     (1 << 2)
    #define PHY_INT_MASK_ANEG_PAGERx_   (1 << 1)


#define PHY_SP_CSR      31
#if 1
#define USE_WAKE_PHY_ON_PCI
#endif


#define LAN9420_MAC_WAKE_MASK (\
                                WAKE_UCAST | WAKE_BCAST | \
                                WAKE_MCAST | WAKE_ARP   | \
                                WAKE_MAGIC \
        )
#define MAC_WAKE_FILTERCMD_ADDR_TYPE_MCAST_    (1 << 3)
#define MAC_WAKE_FILTERCMD_EN_                  (1 << 0)

#define LAN9420_BCAST_WAKE_CRC  0x8029
#define LAN9420_MCAST_WAKE_CRC  0x851F
#define LAN9420_ARP_WAKE_CRC    0x614e
//these precomputed values are derived from 
//CalculateCrc16(bBast, BCAST_LEN, 0)
//CalculateCrc16(bMcast, MCAST_LEN, 0)
//CalculateCrc16(bArp, ARP_LEN, 0)
//where :
//
//#define BCAST_LEN   6
//#define MCAST_LEN   3
//#define ARP_LEN     2
//static const     u8 bBcast[BCAST_LEN] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
//static const     u8 bMcast[MCAST_LEN] = {0x01, 0x00, 0x5E};
//static const     u8 bArp[ARP_LEN] = {0x08, 0x06};

////////////////////////////////////////////////////////////
// local types
////////////////////////////////////////////////////////////

typedef ulong spinlock_int_t;

typedef struct _WAKEUP_FILTER {
    u32 dwFilter0Mask   __attribute__ ((packed));
    u32 dwFilter1Mask   __attribute__ ((packed));
    u32 dwFilter2Mask   __attribute__ ((packed));
    u32 dwFilter3Mask   __attribute__ ((packed));
    u32 dwCommand        __attribute__ ((packed));
    u32 dwOffset        __attribute__ ((packed));
    u32 dwCRC_10        __attribute__ ((packed));
    u32 dwCRC_32        __attribute__ ((packed));
} WAKEUP_FILTER, *PWAKEUP_FILTER;

//9420 has 32bit DMA, in order for it to work
//on 64 bit platforms, we should ensure that 
//DMA descriptor members are 32 bit wide 
//and packed
typedef struct _SMSC9420_DMA_DESCRIPTOR {
    volatile u32 dwStatus    __attribute__ ((packed));
    volatile u32 dwLength    __attribute__ ((packed));
    volatile u32 dwBuffer1   __attribute__ ((packed));
    volatile u32 dwBuffer2   __attribute__ ((packed));
} 
SMSC9420_DMA_DESCRIPTOR, *PSMSC9420_DMA_DESCRIPTOR;


//this structure does not require any unnatural alignment
typedef struct _INFO_RX_SkBUFFERS{
    struct sk_buff * RxSkb;
    dma_addr_t   mapping_RxSkBuffers;
} INFO_RX_SkBUFFERS, *PINFO_RX_SkBUFFERS;

//this structure does not require any unnatural alignment
typedef struct _INFO_TX_SkBUFFERS{
    struct sk_buff * TxSkb;
    dma_addr_t   mapping_TxSkBuffers;
    ulong len;
} INFO_TX_SkBUFFERS, *PINFO_TX_SkBUFFERS;



typedef struct
{
    ulong err_cnt;
    ulong init_err_cnt;
    ulong run_err_cnt;
    ulong last_err_line;
}
gl_drv_error_t;

typedef struct
{
    ulong int_cnt;
    ulong int_self;
    ulong int_handled;
    ulong int_rx;
    ulong int_tx;
    ulong int_err;
}
drv_int_stats_t;

typedef struct
{

}
drv_entry_t;
typedef struct
{

}
drv_exit_t;
typedef struct
{
    drv_entry_t entry;
    drv_exit_t  exit;
}
gl_drv_stats_t;

typedef struct 
{
    gl_drv_stats_t gl_stats;
    gl_drv_error_t gl_error;
    drv_int_stats_t int_stat;
}
drv_accounting_t;

//private data is crafted to 
//have higher aligned members at the 
//begining and lower aligned members towards the end
//most frequently used members are placed close to the top
//perhaps, these frequently accessed members should be regrouped
//to enforce higher alignment on them. this may improve performance 
//on some platforms
typedef struct _PRIVATE_DATA {

    PSMSC9420_DMA_DESCRIPTOR RxDescRing;
    PSMSC9420_DMA_DESCRIPTOR TxDescRing;

    dma_addr_t   rx_dma_addr;
    dma_addr_t   tx_dma_addr;

    ulong        rx_ring_size;
    ulong        tx_ring_size;

    PINFO_TX_SkBUFFERS TxSkBuffers; 
    PINFO_RX_SkBUFFERS RxSkBuffers;

    ulong                LanBase;
    struct pci_dev     *pdev;

    ulong TxHeadIndex, TxTailIndex;
    ulong RxHeadIndex, RxTailIndex;
    ulong TxFull;


    struct net_device *dev;
    
    ulong dwIdRev;
    ulong dwFpgaRev;

    spinlock_t INT_CTL_lock; //this lock is used to protect access to INT_CTL
    spinlock_t INT_CFG_lock; //this lock is used to protect access to INT_CFG

    spinlock_t phy_lock;
    struct tasklet_struct  Smsc9420_PHYTasklet;


    #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
    struct napi_struct napi;
    #endif
    
    //NAPI 
    ulong RxWorkLimit;

    ulong IntDeas_10us;

    struct net_device_stats stats;

    struct timer_list LinkPollingTimer;
    ulong StopLinkPolling;

    
    ulong SoftwareInterruptSignal;
    ulong LanInitialized;
    
    ulong RxRingFull;

    ulong dwSignature;//Should always be set to SMSC9420_PRIVATE_DATA_SIGNATURE
    ulong UseRxCsumHW;
    ulong isr_try_cnt;
    ulong clr_deas_timer_on_rx;
    
    ulong NeedsWake;
    ulong WolWakeupOpts; 
    drv_accounting_t accounting;
    u32   debug_mode;
    u32   GpoSetting;

    u32   dwPhyAddress;
    u32   Phy_OUI;
    u32   Phy_Rev;
    u32   Phy_Model;
    u32   dwLinkSpeed;
    u32   dwLinkSettings;

    u32   RxFoxFilter;
    u32   RxFoxHashLo;
    u32   RxFoxHashHi;

    u32   MacHigh16;
    u32   MacLow32;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
    u32   PCI_conf_copy[64/sizeof(u32)];
#endif
    u16 wLastADV;
    u16 wLastADVatRestart;
} PRIVATE_DATA, * pPD;




typedef enum
{
    enmAC_TYPE_NULL=0,
    enmAC_TYPE_ISR,
    enmAC_TYPE_RX_TASKLET,
    enmAC_TYPE_TX_TASKLET,
    enmAC_TYPE_RX_POLL,
    enmAC_TYPE_INIT,
    enmAC_TYPE_POWER,
    enmAC_TYPE_LAST
}
acc_target_t;

typedef enum
{
    enmAC_STAT_OK=0,
    enmAC_STAT_BA,      //bad arguments
    enmAC_STAT_BI,      //invalid interrupt
    enmAC_STAT_ST,      //wrong state
    enmAC_STAT_BC,      //bad context
    enmAC_STAT_HW_FLT,  //hw flt
    enmAC_STAT_TO,      //timeout
    enmAC_STAT_BM,      //bad memory
    enmAC_STAT_UNKNOWN,
    enmAC_STAT_LAST
}
acc_status_t;
typedef struct
{
    pPD pd;
    acc_status_t  stat;
    acc_target_t  type;
    char *function;
    ulong         line;

}
acc_tocken_t;

    

////////////////////////////////////////////////////////////
//// local data
////////////////////////////////////////////////////////////

static const char *SMSC_CHIP_NAMES[] = {
        "LAN9420"
    };

static const struct pci_device_id Smsc9420_id_table[] = {
    { 
        PCI_VENDOR_ID_9420ESS,                  // Vendor ID
        PCI_DEVICE_ID_9420ESS,                  // Device ID
        PCI_ANY_ID,                             // SubVendor ID
        PCI_ANY_ID,                             // SubDevice ID
                                    
    },
    { 0, }
};

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18))
MODULE_DEVICE_TABLE(pci, Smsc9420_id_table);
#endif




MODULE_LICENSE("GPL");


/////////////////////////////////////////////////////////
// module parameters
/////////////////////////////////////////////////////////
//
// The key parameters for the driver which define performance are:
// tx_ring_size, rx_ring_size and int_deas.
// tx_ring_size: has to be a power of 2 value. 
// apparently should not be > 32. 
// It appears, that linux will not give up 64 or, much over 32 buffers at one
// time. This means that we can not possibly fill large tx ring to its fullest under 
// Linux. Filling the ring can be triggered by udp netperf with mtu sized buffers on 
// fast machine. It would take > 100us to send out each packet. Since UDP traffic does not
// require acks, CPU will be able to provide back-to-back buffers at rates higher than the wire speed.
// On larger tx rings, OS will stop calling hard_start_xmit() even though the queue is still up.
// 
// int_deas: is a value in 10us units, which defines for how long interrupt will be deasserted.
// Assuming a value of 50, interrupt are deasserted for 500us. 
// In case of minimum sized frames (ETH_ZLEN) it is possible to recieve over 100 frames within
// interrupt deassertion period. Although, this case is artificial, it is a good idea to keep 
// rx dma ring large. 128 is probably a good value. rx_ring_size has to be a power of 2 value,
// 256 is probably too large for default deassertion timer.
// Decreasing deasertion interval may hurt tx performance. Consider TXing MTU sized buffers.
// It will take over 500us to TX 5 buffers. Therefore, 500us int_deas would prevent the DMA from 
// emptying TX dma ring. We would end up filling the ring and interrupting every 3-4 packets.
// Shortening interrupt deasertion interval  will make situation even worse.
// On the other hand, large deassertion interval will increase rx_ring_size. It will also penalize non bulk 
// recieves. Such, as a single rx packet would have to be delayed in delivery for up to deassertion interval.
// It appears that default values of:
// rx_ring_size = 128 (256 on slower CPUs)
// int_deas = 50
// tx_ring_size = 32 or 16
// Look optimal for general purpose applications on all platforms.
///////////////////////////////////////////////////////////////
/// General design notes
///////////////////////////////////////////////////////////////
// 1. Tx path:
// Buffer recycling happens in a context of hard_start_xmit()
// Prior to placing a new buffer in the ring.
// Assuming everything works out good for us, no TX interrupt is needed.
// If however, we start filling up the TX ring, we stop the queue before
// taking the last descriptor. We also, instruct DMA to issue a TX interrupt
// on completion. Tx ISR wakes the queue. 
// It is important to know, that queue should only be stoped in hard_start_xmit(),
// it should only be woken in  Tx ISR.
// All other accesses to queue will result in packets dropped, or lock ups!!
// TX interrupt is never masked or disabled. 
// It has been observed on 2.6.23 and 2.6.26 x86 SMP and x86_64 SMP  platforms, that
// in case of large packet back-to-back Tx, CPU fills up the Tx DMA ring. This is normal.
// Because, it takes over 100us to send out MTU sized buffer, recycler does not have any 
// buffers to recycle in back to back Tx case. It may be possible for OS to provide 32 buffers
// within 100us.
// However, it has been noted that with ring size 64 and higher, OS stops calling hard_start_xmit()
// after providing ~32 buffers. This leads to a natural deadlock. hard_start_xmit() awaits more
// buffers to transmit. OS awaits it's buffers to get recycled. Since, hard_start_xmit() is the
// only "recycler", not calling hard_start_xmit() never triggers a recycler. 
// 
// 2. Rx path:
// Only napi mode is supported. Poll function should operate with RX interrupts disabled.
// Poll function is the only place to enable RX interrupts upon Poll completion.
// RX ISR is the only place to disable RX interrupts.
// Because, DMA_INT_ENA is not thread safe, we should use a lock to perform read-modify-write
// on RX_ENA_, TX_ENA_ bits. However, since we know that TX interrupts are never disabled,
// And Poll and Rx ISR are in producer-consumer relationship, we can avoid spinlocking on
// DMA_INT_ENA access.
// Additionally, it is important to make sure that Rx and Tx softirq contexts do not share any
// data. This is currently the case. On SMP systems it is possible to have Poll and hard_start_xmit()
// Executing simultaneously on different CPUs. Avoiding shared data between Tx and Rx eliminates all the
// locks in the main datapath.
// Note: it is possible for Poll to get preemted under certain conditions. In particular, it has been 
// observed on 200Mhz ARM926-ejs under 2.6.9 kernel, that netif_receive_skb() may yield to others
// including hard_start_xmit().
// This is a special case of napi performance, which is important to keep in mind. 
// 
// 3. start/stop/suspend/resume only access DMA_INT_ENA with all interrupts disabled. 
// The access to INT_CTL is protected with a spinlock. Since, start/stop/suspend/resume are not in the
// main datapath, there is no reccuring penalty.
// INT_CFG contains Interrupt_Deassertion parameters. Poll, rearms interrupt deassertion interval before
// enabling an interrupt. spinlock is used to protect INT_CFG. However, there is noone to compete for this
// lock in the main datapath. There is no reccuring penalty.
// It is unclear if setting INT_DEAS_CLR in Poll is beneficial for both Tx and Rx path.
// It may be reasonable to keep the interval detached from Poll. 
// 
///////////////////////////////////////////////////////////////

ulong debug_mode=(
        DBG_MCAST       |
        DBG_TRACE       |
        DBG_WARNING     |
        DBG_INIT        |
        DBG_CLOSE       |
        DBG_LINK        |
        DBG_IOCTL       |
#if 0
        DBG_PWR         |
        DBG_ETH_GPIO    |
#endif
        0
        );
    module_param(debug_mode, ulong, 0);
MODULE_PARM_DESC(debug_mode,"bit 0 enables trace points, bit 1 enables warning points, bit 2 enables eth gpios, bit 3 enables gen gpios");

ulong int_deas=DEFAULT_INT_DEAS_10us;
    module_param(int_deas, ulong, 0);
MODULE_PARM_DESC(int_deas,"Interrupt Deassertion Interval in 10us units");

ulong link_mode=0x7FUL;
    module_param(link_mode, ulong, 0);
MODULE_PARM_DESC(link_mode,"Link Mode");

ulong mac_addr_hi16=0xFFFFFFFFUL;
    module_param(mac_addr_hi16, ulong, 0);
MODULE_PARM_DESC(mac_addr_hi16,"Specifies the high 16 bits of the mac address");

ulong mac_addr_lo32=0xFFFFFFFFUL;
    module_param(mac_addr_lo32, ulong, 0);
MODULE_PARM_DESC(mac_addr_lo32,"Specifies the low 32 bits of the mac address");

ulong napi_weight=64UL;

    module_param(napi_weight, ulong, 0);
MODULE_PARM_DESC(napi_weight,"The weight value to use if NAPI is used");

ulong rx_ring_size=DEFAULT_RX_RING_SIZE;
    module_param(rx_ring_size, ulong, 0);
MODULE_PARM_DESC(rx_ring_size,"Specifies the size of the rx ring by number of descriptors");

ulong tx_ring_size=DEFAULT_TX_RING_SIZE;
    module_param(tx_ring_size, ulong, 0);
MODULE_PARM_DESC(tx_ring_size,"Specifies the size of the tx ring by number of descriptors. Has to be a \
        power of 2");
int rx_Csum=1;
    module_param(rx_Csum,bool, 0);
MODULE_PARM_DESC(rx_Csum,"Enable Rx Hardware Checksum Offload");

ulong isr_try_cnt=DEFAULT_ISR_TRY_CNT;
    module_param(isr_try_cnt,ulong, 0);

MODULE_PARM_DESC(isr_try_cnt, "Number of extra passes in ISR. Set to\
            0 for no isr spinning"); 

int rst_timer_on_rx = 1;
    module_param(rst_timer_on_rx, bool, 0);
MODULE_PARM_DESC(rst_timer_on_rx, "If set: deassertion timer starts again after \
        napi POLL is finished. If cleared deasertion timer is at free run");





/////////////////////////////////////////////////////////
// local prototypes
/////////////////////////////////////////////////////////


static void Rx_hw_stop(pPD pd);
static void Rx_hw_start(pPD pd);
static ulong pd_set_ids(pPD pd);
static int Smsc9420_probe(struct pci_dev * pdev, const struct pci_device_id * id);
static void Smsc9420_remove(struct pci_dev *pdev);
static void SMSC9420_parse_module_parameters(void);
static int Smsc9420_open(struct net_device *dev);
static int Smsc9420_stop(struct net_device *dev);
static void smsc9420_down(struct net_device *ndev);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
irqreturn_t Smsc9420_ISR(int irq,void *dev_id,struct pt_regs *regs);
#else
irqreturn_t Smsc9420_ISR(int irq,void *dev_id);
#endif
static void smsc9420_pci_push(pPD pd);
static void get_set_mac_addr(pPD pd);
static int Smsc9420_hard_start_xmit(struct sk_buff *skb, struct net_device *dev);
static ulong put_new_tx_dma(pPD pd, struct sk_buff* skb);
static void tx_eval_status(pPD pd, const ulong status, const ulong index);
static void recycle_tx_dma(pPD pd);
static struct net_device_stats * Smsc9420_get_stats(struct net_device * const dev);
static void Smsc9420_set_multicast_list(struct net_device *dev);
static int Smsc9420_ethtool_ioctl(pPD pd, void __user * userAddr);
static int Smsc9420_do_ioctl(
    struct net_device *dev,
    struct ifreq *ifr,
    int cmd);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
static int Smsc9420_rx_poll(struct net_device *dev, int * budget);
#else
static int Smsc9420_rx_poll(struct napi_struct *napi, int budget);
#endif
static void rx_set_new_buffer(pPD pd, const ulong index);
static void rx_reuse_old_buffer(pPD pd, const ulong index);
static void Smsc9420_PHYTaskletRoutine(ulong data);


static ulong Lan_Initialize(pPD pd);
static void pmt_rst_phy(pPD pd);
static void rst_mac_dmac(pPD pd);
static void config_gpio(pPD pd);
static void Lan_SignalSoftwareInterrupt(pPD pd);
//int handlers
static void Lan_HandleGPTInt(pPD pd);
static void Lan_HandleSWInt(pPD pd);
static void Lan_HandleSBErrInt(pPD pd);
static void Lan_HandleMBErrInt(pPD pd);
static void Lan_HandlePHYInt(pPD pd);
static void Lan_HandleWAKEInt(pPD pd);
static void Lan_HandleGPIOInt(pPD pd);
static void Lan_HandleDMAInt(pPD  pd);

static void Lan_EnableInterrupt(pPD pd,u32 dwIntEnMask);
static void Lan_DisableInterrupt(pPD pd,u32 dwIntEnMask);
static void Lan_EnableIRQ(pPD pd);
static void Lan_DisableIRQ(pPD pd);
static void Lan_SetIntDeas(pPD pd,u32 dwIntDeas);
static void Lan_ClearDeassertionTimer(pPD pd);
//this one uses nolock
static void Lan_Disable_RXDMACInt_nolock(pPD pd);
//this one uses nolock
static void Lan_Enable_RXDMACInt_nolock(pPD pd);
static void Lan_Disable_TXDMACInt_nolock(pPD pd);
static void Lan_Enable_TXDMACInt_nolock(pPD pd);


static u16 Phy_GetRegW(
    pPD pd,
    u32 dwRegIndex);
static void Phy_SetRegW(
    pPD pd,
    u32 dwRegIndex,u16 wVal);
static ulong Phy_Initialize(pPD pd, u32 dwLinkRequest);
static void Phy_SetLink(pPD pd,
                 u32 dwLinkRequest);
static void Phy_CheckLink(ulong ptr);
static void Phy_UpdateLinkMode(pPD pd);
static void Phy_GetLinkMode(
    pPD pd);

static ulong Tx_Initialize(pPD pd);
static void Tx_hw_stop(pPD pd);
static void Tx_hw_start(pPD pd);
static void Tx_CleanUp(pPD pd);
static void Tx_buffers_destroy(pPD pd);


static ulong Rx_Initialize(pPD pd);
static void Rx_CleanUp(pPD pd);
static ulong Rx_IncrementHeadIndex(pPD pd);
static ulong Rx_IncrementTailIndex(pPD pd);
static void Rx_buffers_destroy(pPD pd);
static ulong Rx_GetFreeDescriptorCount(pPD pd);
static u32 Rx_Hash(u8 addr[6]);
static void Rx_SetMulticastList(pPD pd);

extern void apm_queue_event(apm_event_t event);

static void set_mac_sleep(pPD pd, const ulong mac_down);
static void set_phy_sleep(pPD pd, const ulong phy_down);
#ifdef CONFIG_PM
static void ResetWakeupEvents(struct net_device *dev);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
    static int Smsc9420_suspend (struct pci_dev *pdev, u32 state);
#else
    static int Smsc9420_suspend (struct pci_dev *pdev, pm_message_t state);
#endif
static int Smsc9420_resume(struct pci_dev *pdev);
static u16 CalculateCrc16(const u8 * bpData,
        const ulong dwLen, const ulong fBitReverse);
static void process_wake_and_hw_sleep(struct net_device *ndev);
static ulong is_link_up(pPD pd);
#endif



static struct pci_driver Smsc9420_driver = {
    .name =         DRV_NAME,
    .id_table =     Smsc9420_id_table,
    .probe =        Smsc9420_probe,
    .remove =       __devexit_p(Smsc9420_remove),
    #ifdef CONFIG_PM
    .suspend    = Smsc9420_suspend,
    .resume     = Smsc9420_resume,
    #endif /* CONFIG_PM */
};


//////////////////////////////////////////////////////////
/// end of declarations
//////////////////////////////////////////////////////////








#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19))
static int
Smsc9420_get_settings(struct net_device *netdev, struct ethtool_cmd *ecmd)
{
    pPD pd=(pPD)netdev_priv(netdev);

    SMSC_TRACE(DBG_INIT,"--> %s()", __FUNCTION__);
    ecmd->supported=
        SUPPORTED_10baseT_Half |
        SUPPORTED_10baseT_Full |
        SUPPORTED_100baseT_Half |
        SUPPORTED_100baseT_Full |
        SUPPORTED_Autoneg |
        SUPPORTED_MII;
    ecmd->advertising=ADVERTISED_MII;
    if(pd->dwLinkSettings & LINK_SPEED_10HD)
        ecmd->advertising|=ADVERTISED_10baseT_Half;
    if(pd->dwLinkSettings & LINK_SPEED_10FD)
        ecmd->advertising|=ADVERTISED_10baseT_Full;
    if(pd->dwLinkSettings & LINK_SPEED_100HD)
        ecmd->advertising|=ADVERTISED_100baseT_Half;
    if(pd->dwLinkSettings & LINK_SPEED_100FD)
        ecmd->advertising|=ADVERTISED_100baseT_Full;
    if(pd->dwLinkSettings & LINK_AUTO_NEGOTIATE) {
        ecmd->advertising|=ADVERTISED_Autoneg;
        ecmd->autoneg=AUTONEG_ENABLE;
    } else ecmd->autoneg=AUTONEG_DISABLE;
    if(pd->dwLinkSpeed & (LINK_SPEED_100HD|LINK_SPEED_100FD))
        ecmd->speed=SPEED_100;
    else ecmd->speed=SPEED_10;
    if(pd->dwLinkSpeed & (LINK_SPEED_10FD|LINK_SPEED_100FD))
        ecmd->duplex=DUPLEX_FULL;
    else ecmd->duplex=DUPLEX_HALF;
    ecmd->port=PORT_MII;
    ecmd->phy_address=(u8)pd->dwPhyAddress;
    ecmd->transceiver=XCVR_INTERNAL;
    ecmd->maxtxpkt=0;
    ecmd->maxrxpkt=0;

	return 0;
}

static int
Smsc9420_set_settings(struct net_device *netdev, struct ethtool_cmd *ecmd)
{
    u16 speed=0;
    u8 duplex=0;
    u8 autoneg=0;
    int result=0;

    pPD pd=(pPD)netdev_priv(netdev);
    SMSC_TRACE(DBG_INIT,"--> %s()", __FUNCTION__);
    
    if(pd->dwLinkSettings&LINK_AUTO_NEGOTIATE) {
        autoneg=AUTONEG_ENABLE;
    } else {
        autoneg=AUTONEG_DISABLE;
    }
    if(pd->dwLinkSpeed&(LINK_SPEED_100HD|LINK_SPEED_100FD))
    {
        speed=SPEED_100;
    } else {
        speed=SPEED_10;
    }
    if(pd->dwLinkSpeed&(LINK_SPEED_10FD|LINK_SPEED_100FD))
    {
        duplex=DUPLEX_FULL;
    } else {
        duplex=DUPLEX_HALF;
    }
    if((ecmd->speed!=100)&&(ecmd->speed!=10)) {
        result=-EOPNOTSUPP;
    }
    if((ecmd->duplex!=DUPLEX_FULL)&&(ecmd->duplex!=DUPLEX_HALF)) {
        result=-EOPNOTSUPP;
    }
    if((ecmd->autoneg!=AUTONEG_ENABLE)&&(ecmd->autoneg!=AUTONEG_DISABLE)) {
        result=-EOPNOTSUPP;
    }
    if((ecmd->autoneg!=autoneg)||
        (ecmd->speed!=speed)||
        (ecmd->duplex!=duplex))
    {
        if(ecmd->autoneg==AUTONEG_ENABLE) {
            u16 wAdvertisement=Phy_GetRegW(pd,PHY_ANEG_ADV);
            wAdvertisement&=(~PHY_ANEG_ADV_SPEED_);
            if(ecmd->speed==SPEED_100) {
                if(ecmd->duplex==DUPLEX_FULL) {
                    wAdvertisement|=PHY_ANEG_ADV_100F_;
                } else {
                    wAdvertisement|=PHY_ANEG_ADV_100H_;
                }
            } else {
                if(ecmd->duplex==DUPLEX_FULL) {
                    wAdvertisement|=PHY_ANEG_ADV_10F_;
                } else {
                    wAdvertisement|=PHY_ANEG_ADV_10H_;
                }
            }
            Phy_SetRegW(pd,PHY_ANEG_ADV,wAdvertisement);
            Phy_SetRegW(pd,PHY_BCR,
                (PHY_BCR_ANEG_EN_|PHY_BCR_RST_ANEG_));
        } else {
            u16 wBcr=Phy_GetRegW(pd,PHY_BCR);
            if(ecmd->speed==SPEED_100) {
                wBcr|=PHY_BCR_SPEED_SELECT_;
            } else {
                wBcr&=(~PHY_BCR_SPEED_SELECT_);
            }
            if(ecmd->duplex==DUPLEX_FULL) {
                wBcr|=PHY_BCR_DUPLEX_MODE_;
            } else {
                wBcr&=(~PHY_BCR_DUPLEX_MODE_);
            }
            Phy_SetRegW(pd,PHY_BCR,wBcr);
        }
    }

	return result;
}
static void
Smsc9420_get_drvinfo(struct net_device *netdev,
                       struct ethtool_drvinfo *drvinfo)
{
    pPD pd=(pPD)netdev_priv(netdev);

    strcpy(drvinfo->driver, DRV_NAME);
    strcpy(drvinfo->bus_info, pci_name(GET_PD_PDEV(pd)));
    strcpy(drvinfo->version, DRV_VERSION);
    return;
}


static void
Smsc9420_get_wol(struct net_device *netdev, struct ethtool_wolinfo *wol)
{
    pPD pd=(pPD)netdev_priv(netdev);

    wol->supported=( WAKE_UCAST | WAKE_BCAST | 
                     WAKE_MCAST | WAKE_ARP | 
                     WAKE_MAGIC |
#ifdef USE_WAKE_PHY_ON_PCI
                     WAKE_PHY   |
#endif
                     0);
    wol->wolopts= GET_PD_WOLOPTS(pd); 
    return;
}

static int
Smsc9420_set_wol(struct net_device *netdev, struct ethtool_wolinfo *wol)
{
    pPD pd=(pPD)netdev_priv(netdev);
    SET_PD_WOLOPTS(pd, wol->wolopts);
	return 0;
}

static uint32_t
Smsc9420_get_msglevel(struct net_device *netdev)
{
    pPD pd = (pPD)netdev_priv(netdev);
	return GET_PD_DEBUG_MODE(pd);
}

static void
Smsc9420_set_msglevel(struct net_device *netdev, uint32_t data)
{
    pPD pd = (pPD)netdev_priv(netdev);
    SET_PD_DEBUG_MODE(pd, data);
    return;
}

static int
Smsc9420_nway_reset(struct net_device *netdev)
{
    pPD pd=(pPD)netdev_priv(netdev);

    SMSC_TRACE(DBG_INIT,"--> %s()", __FUNCTION__);
    Phy_SetRegW(pd,PHY_BCR,
        (PHY_BCR_ANEG_EN_|PHY_BCR_RST_ANEG_));
	return 0;
}


static uint32_t
Smsc9420_get_rx_csum(struct net_device *netdev)
{
	return 0;
}


static uint32_t
Smsc9420_get_tx_csum(struct net_device *netdev)
{
	return 0;
}


static const struct ethtool_ops Smsc9420_ethtool_ops = {
	.get_settings           = Smsc9420_get_settings,
	.set_settings           = Smsc9420_set_settings,
	.get_drvinfo            = Smsc9420_get_drvinfo,
	.get_wol                = Smsc9420_get_wol,
	.set_wol                = Smsc9420_set_wol,
	.get_msglevel           = Smsc9420_get_msglevel,
	.set_msglevel           = Smsc9420_set_msglevel,
	.nway_reset             = Smsc9420_nway_reset,
	.get_link               = ethtool_op_get_link,
	.get_rx_csum            = Smsc9420_get_rx_csum,
	.get_tx_csum            = Smsc9420_get_tx_csum,
	.get_sg                 = ethtool_op_get_sg,
	.set_sg                 = ethtool_op_set_sg,
	.get_tso                = ethtool_op_get_tso,
};

void Smsc9420_set_ethtool_ops(struct net_device *netdev)
{
	SET_ETHTOOL_OPS(netdev, &Smsc9420_ethtool_ops);
}
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)) */


static void SMSC9420_parse_module_parameters(void)
{

    SMSC_TRACE(DBG_INIT,"Driver Parameters");
    SMSC_TRACE(DBG_INIT,"  lan_base  will be auto-detected");
    SMSC_TRACE(DBG_INIT,    "  irq  will be auto-detected");

    if(int_deas>0xFF) {
        int_deas=30;
        SMSC_WARNING("int_deas is out of range, it has been reset to %ld",int_deas);
    }
    SMSC_TRACE(DBG_INIT,    "  int_deas         = %ld",int_deas);
    SMSC_TRACE(DBG_INIT,    "  link_mode        = 0x%02lX",link_mode);
    if(mac_addr_hi16==0xFFFFFFFFUL) {
        SMSC_TRACE(DBG_INIT,"  mac_addr_hi16    = 0x%08lX, will attempt to read from LAN9420",mac_addr_hi16);
        SMSC_TRACE(DBG_INIT,"  mac_addr_lo32    = 0x%08lX, will attempt to read from LAN9420",mac_addr_lo32);
    } else {
        if(mac_addr_hi16&0xFFFF0000UL) {
            //The high word is reserved
            SMSC_WARNING("mac_addr_hi16 = 0x%08lX, reserved bits are high.",mac_addr_hi16);
            mac_addr_hi16&=0x0000FFFFUL;
            SMSC_WARNING("reseting to mac_addr_hi16 = 0x%08lX",mac_addr_hi16);
        }
        if(mac_addr_lo32&0x00000001UL) {
            //bit 0 is the I/G bit
            SMSC_WARNING("mac_addr_lo32 = 0x%08lX, I/G bit is set.",mac_addr_lo32);
            mac_addr_lo32&=0xFFFFFFFEUL;
            SMSC_WARNING("reseting to mac_addr_lo32 = 0x%08lX",mac_addr_lo32);          
        }
        SMSC_TRACE(DBG_INIT,"  mac_addr_hi16    = 0x%08lX",mac_addr_hi16);
        SMSC_TRACE(DBG_INIT,"  mac_addr_lo32    = 0x%08lX",mac_addr_lo32);
    }
    SMSC_TRACE(DBG_INIT,    "  debug_mode       = 0x%08lX %s %s %s %s",debug_mode,
        debug_mode&DBG_TRACE?"(TRACE)":"",
        debug_mode&DBG_WARNING?"(WARNING)":"",
        debug_mode&DBG_ETH_GPIO?"(ETH_GPO)":"",
        debug_mode&DBG_SYS_GPIO?"(SYS_GPO)":"");
    if(rx_ring_size<MIN_RX_RING_SIZE) {
        SMSC_WARNING("rx_ring_size is too small");
        SMSC_WARNING("resetting to rx_ring_size=%lu", MIN_RX_RING_SIZE);
        rx_ring_size=MIN_RX_RING_SIZE;
    }
    if(rx_ring_size>MAX_RX_RING_SIZE) {
        SMSC_WARNING("rx_ring_size is too large");
        SMSC_WARNING("resetting to rx_ring_size=%lu", MAX_RX_RING_SIZE);
        rx_ring_size=MAX_RX_RING_SIZE;
    }

    if((rx_ring_size - 1) & (rx_ring_size))
    {
        //rx_ring_size is not a power of 2
        SMSC_WARNING("rx_ring_size is not a power of 2");
        SMSC_WARNING("resetting to default value: %lu", DEFAULT_RX_RING_SIZE);
        rx_ring_size = DEFAULT_RX_RING_SIZE;
    }
    SMSC_TRACE(DBG_INIT,"  rx_ring_size     = %ld",rx_ring_size);
    //napi_weight should not exceed  rx_ring_size
    //on the other hand, it is unclear if there are benefits
    //of having napi_weight  below rx_ring_size
    //one can certainly try to exepriment with lower napi_weights
    //with 1 being an abolsute minimum... interrupt on each packet.
    //0 value is probably illegal...
    if(napi_weight > rx_ring_size)
    {
        SMSC_WARNING("napi_weight: %ld exceeds rx_ring_size: %ld",
                napi_weight,rx_ring_size);
        napi_weight = rx_ring_size/DEFAULT_RX_RING_TO_NAPI_WEIGHT_FACTOR;
        SMSC_WARNING("resetting napi_weight to 1/%lu rx_ring_size: %ld",
                DEFAULT_RX_RING_TO_NAPI_WEIGHT_FACTOR, napi_weight);

    }
    else if(!napi_weight)
    {
            SMSC_WARNING("napi_weight: is zero");
            napi_weight = rx_ring_size / DEFAULT_RX_RING_TO_NAPI_WEIGHT_FACTOR;
            SMSC_WARNING("resetting napi_weight to : %lu",
                napi_weight);
    }
    else
    {
        SMSC_TRACE(DBG_INIT,"  napi_weight      = %lu",napi_weight);
    }
    if(tx_ring_size<MIN_TX_RING_SIZE) {
        SMSC_WARNING("tx_ring_size is too small");
        SMSC_WARNING("resetting to tx_ring_size=%lu",MIN_TX_RING_SIZE);
        tx_ring_size=MIN_TX_RING_SIZE;
    }

    if(tx_ring_size>MAX_TX_RING_SIZE) {
        SMSC_WARNING("tx_ring_size is too large");
        SMSC_WARNING("resetting to tx_ring_size=%lu", MAX_TX_RING_SIZE);
        tx_ring_size=MAX_TX_RING_SIZE;
    }
    if((tx_ring_size - 1) & (tx_ring_size))
    {
        //tx_ring_size is not a power of 2
        SMSC_WARNING("tx_ring_size is not a power of 2");
        SMSC_WARNING("resetting to default value: %lu", DEFAULT_TX_RING_SIZE);
        tx_ring_size = DEFAULT_TX_RING_SIZE;
    }
    
    SMSC_TRACE(DBG_INIT,"  tx_ring_size     = %ld",tx_ring_size);


    if(rx_Csum) {
        SMSC_TRACE(DBG_INIT,"  Rx HW Checksum Enabled");
    } else {
        SMSC_TRACE(DBG_INIT,"  Rx HW Checksum Disabled");
    }
    if(isr_try_cnt > MAX_ISR_TRY_CNT)
    {
        isr_try_cnt = DEFAULT_ISR_TRY_CNT;
        SMSC_TRACE(DBG_INIT,"  ISR spin count limit exceeds max value of %lu\n",
                MAX_ISR_TRY_CNT);
        SMSC_TRACE(DBG_INIT,"  resetting to default value of %lu\n",
                DEFAULT_ISR_TRY_CNT);
    }


    return;
    
}


static int __devinit 
Smsc9420_probe(struct pci_dev * pdev, 
        const struct pci_device_id 
        __attribute__ ((unused))
        *id)

{
    unsigned short tmp16;
    unsigned char tmp8;
    struct net_device *ndev;
    pPD     ndev_priv;
    int result=0;
    ulong virt_addr;
    
    atomic_t tmp = ATOMIC_INIT(1);

    SMSC_TRACE(DBG_INIT,"--> Smsc9420_probe()");
    SMSC_TRACE(DBG_INIT,"LAN9420 Driver Version: %s",
            DRV_VERSION);
    SMSC_TRACE(DBG_INIT,"Compiled: %s, %s",__DATE__,__TIME__);


    SMSC9420_parse_module_parameters();

    /* First do the PCI initialisation */
    result=pci_enable_device(pdev);
    if (unlikely(result))
    {
        printk(KERN_ERR 
                "Cannot enable SMSC9420 board, aborting");
        return result;
    }

    SMSC_TRACE(DBG_INIT,"irq= 0x%08x", pdev->irq);

    pci_set_master(pdev); 

    //attempt to allocate net_device with pd
    //there is no need to call ether_setup()
    //alloc_etherdev does all the default setups
    ndev = alloc_etherdev(sizeof(*ndev_priv));
    //fail early 
    if(!ndev)
    {
        printk(KERN_ERR "ether device alloc failed, aborting\n");
        return -ENOMEM;
    }
    //register parent in ndev
    SET_NETDEV_DEV(ndev, &pdev->dev);



    /* clear PCI status, if any */ 
    pci_read_config_word (pdev,PCI_STATUS, &tmp16);
    

    pci_read_config_byte (pdev, PCI_LATENCY_TIMER,&tmp8);
    SMSC_TRACE(DBG_INIT,"Latency Timer = %02x", tmp8);


    pci_read_config_word (pdev,PCI_DEVICE_ID, &tmp16);
    SMSC_TRACE(DBG_INIT,"deviceId= 0x%04x", tmp16);
    
    if(!(pci_resource_flags(pdev, BAR3) & IORESOURCE_MEM)) {
        SMSC_WARNING("Cannot find proper PCI device base address, aborting.\n");
        goto err_out_free_dev;
    }

    if((pci_request_regions(pdev,DRV_NAME))) {
        SMSC_WARNING("Cannot obtain PCI resources, aborting.\n");
        goto err_out_disable_pdev;
    }
    if(pci_set_dma_mask(pdev, DMA_32BIT_MASK)) {
        SMSC_WARNING("No usable DMA configuration, aborting.\n");
        goto err_out_free_region;
    }
    virt_addr = (ulong)ioremap(pci_resource_start(pdev, BAR3),
                        pci_resource_len(pdev, BAR3));

    //here we map registers based on CPU's endiannes
    //registers are double mapped with 0 offset for LE and 200h for BE
    if(!virt_addr)
    {
        SMSC_WARNING("Cannot map device registers, aborting.\n");
        goto err_out_free_res;
    }

    virt_addr += LAN9420_CPSR_ENDIAN_OFFSET;

    ndev->base_addr = (ulong)virt_addr;

    SMSC_TRACE(DBG_INIT,"lan_base= 0x%08lx", (ulong)virt_addr);

    //init net_device private 
    ndev_priv = netdev_priv(ndev);

    ndev_priv->rx_ring_size = rx_ring_size+1;   //we add one, to ease head/tail index handling 
    ndev_priv->tx_ring_size = tx_ring_size;   


    //pci descriptors are created in the PCI consistent area
    ndev_priv->RxDescRing = pci_alloc_consistent(pdev,
            sizeof(SMSC9420_DMA_DESCRIPTOR) * ndev_priv->rx_ring_size
            +
            sizeof(SMSC9420_DMA_DESCRIPTOR) * ndev_priv->tx_ring_size,
            &ndev_priv->rx_dma_addr);

    if(!ndev_priv->RxDescRing)
    {
        goto err_out_free_res;
    }
    //descriptors are aligned due to the nature of pci_alloc_consistent

    ndev_priv->TxDescRing= (PSMSC9420_DMA_DESCRIPTOR)
        (ndev_priv->RxDescRing + ndev_priv->rx_ring_size);
    ndev_priv->tx_dma_addr = ndev_priv->rx_dma_addr + sizeof(SMSC9420_DMA_DESCRIPTOR)*
        ndev_priv->rx_ring_size;

//set default PD_values
    //set PD parent
    SET_PD_PDEV(ndev_priv, pdev);
    //set PD own device
    SET_PD_DEV(ndev_priv, ndev);

    SET_PD_LANBASE(ndev_priv, virt_addr);
    SET_PD_WOLOPTS(ndev_priv, SMSC9420_DEFAULT_WOL_OPTS);

    SET_PD_INTDEAS(ndev_priv, int_deas);

    SET_PD_DEBUG_MODE(ndev_priv,debug_mode);
    SET_PD_ISR_CNT(ndev_priv,isr_try_cnt);
    if(rst_timer_on_rx)
    {
        SET_CLR_TIMER_ON_RX(ndev_priv);
    }
    else
    {
        RST_CLR_TIMER_ON_RX(ndev_priv);
    }

    //setup tasklet(s)

    ndev_priv->Smsc9420_PHYTasklet.next = NULL;
    ndev_priv->Smsc9420_PHYTasklet.state= 0;
    memcpy(&ndev_priv->Smsc9420_PHYTasklet.count, &tmp, sizeof(atomic_t));
    ndev_priv->Smsc9420_PHYTasklet.func= Smsc9420_PHYTaskletRoutine;
    ndev_priv->Smsc9420_PHYTasklet.data= (ulong)ndev_priv;

    SET_PD_RxCSUM(ndev_priv,rx_Csum);
    if(pd_set_ids(ndev_priv))
    {
        //destroy all (free_netdev(ndev))
        goto err_out_free_desc;

    }
    //this is important to set MAC address here 
    //newer kernel wouldn't open without  it.
    //however, because we will reset MAC, 
    //we should record this mac address in pd, and
    //rewrite it in MAC after reset.
    //additionally, we should make sure, PHY is not in power down mode now.
    //to do this, let's rst phy.... and mac 

    pmt_rst_phy(ndev_priv);
    rst_mac_dmac(ndev_priv);
    get_set_mac_addr(ndev_priv);
    strcpy(ndev->name,"eth%d");
    ndev->open = Smsc9420_open;
    ndev->stop = Smsc9420_stop;
    ndev->hard_start_xmit=Smsc9420_hard_start_xmit;
    ndev->get_stats=Smsc9420_get_stats;
    ndev->do_ioctl=Smsc9420_do_ioctl;

    #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19))
    Smsc9420_set_ethtool_ops(ndev);
    #endif

    ndev->set_multicast_list=Smsc9420_set_multicast_list;
    ndev->irq = pdev->irq;

    
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
     ndev->poll=Smsc9420_rx_poll;
     ndev->weight=napi_weight;
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
     netif_napi_add(ndev, &ndev_priv->napi, Smsc9420_rx_poll, napi_weight);
#endif
    /* ack any pending wake events, disable PME */
    if(pci_enable_wake(pdev, 0, 0))
	SMSC_WARNING("Error clearing wake event\n");

    result=register_netdev(ndev);

    pci_set_drvdata(pdev, ndev);

    
    if(result) {
        SMSC_WARNING("error %i registering device",result);
        goto err_out_free_desc;
    } else {
        SMSC_TRACE(DBG_INIT,"  Interface Name = \"%s\"",ndev->name);
    }

    spin_lock_init(&(ndev_priv->INT_CTL_lock)); 
    spin_lock_init(&(ndev_priv->INT_CFG_lock));
    spin_lock_init(&(ndev_priv->phy_lock));
    //has to happen after phy_lock is ready
    set_phy_sleep(ndev_priv, 1);

    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
    SET_MODULE_OWNER(dev);
    #endif

    SMSC_TRACE(DBG_INIT,"<-- Smsc9420_init_module()");
    return 0;

err_out_free_desc:
    pci_free_consistent(pdev, 
            sizeof(SMSC9420_DMA_DESCRIPTOR) * (rx_ring_size + tx_ring_size) ,
            ndev_priv->RxDescRing,
            ndev_priv->rx_dma_addr);
err_out_free_res:
      iounmap((void __iomem *)(virt_addr-LAN9420_CPSR_ENDIAN_OFFSET));
//    pci_iounmap(pdev, virt_addr);
err_out_free_region:
    pci_release_regions(pdev);
err_out_disable_pdev:
    pci_disable_device(pdev);
err_out_free_dev:
    free_netdev(ndev);
    pci_set_drvdata(pdev, NULL);

    return -ENODEV;
}


static void get_set_mac_addr(pPD pd)
{
    u32 u32High16=0;
    u32 u32Low32=0;
    struct net_device *dev;

    if(unlikely(!pd))
    {
        return;
    }
     dev = GET_PD_DEV(pd);
     if(mac_addr_hi16==0xFFFFFFFF) {
            u32High16=Lan_GetRegDW(ADDRH);
            u32Low32=Lan_GetRegDW(ADDRL);
            if((u32High16==0x0000FFFFUL)&&(u32Low32==0xFFFFFFFF))
            {
                pd->MacHigh16=0x00000070;
                pd->MacLow32=0x110F8000;

                SMSC_TRACE(DBG_INIT, "Mac Address is set by default to 0x%04lX%08lX",
                   (ulong)u32High16,(ulong)u32Low32);
            } else {
                SMSC_TRACE(DBG_INIT, "Mac Address is read from LAN9420 as 0x%04lX%08lX",
                    (ulong)u32High16,(ulong)u32Low32);
                pd->MacHigh16 = u32High16;
                pd->MacLow32  = u32Low32;
            }
        } else {
            u32High16=mac_addr_hi16;
            u32Low32=mac_addr_lo32;

            pd->MacHigh16 = u32High16;
            pd->MacLow32  = u32Low32;
            SMSC_TRACE(DBG_INIT, "Mac Address is set by parameter to 0x%04lX%08lX",
                    (ulong)u32High16,(ulong)u32Low32);
        }

        dev->dev_addr[0]=LOBYTE(LOWORD(pd->MacLow32));
        dev->dev_addr[1]=HIBYTE(LOWORD(pd->MacLow32));
        dev->dev_addr[2]=LOBYTE(HIWORD(pd->MacLow32));
        dev->dev_addr[3]=HIBYTE(HIWORD(pd->MacLow32));
        dev->dev_addr[4]=LOBYTE(LOWORD(pd->MacHigh16));
        dev->dev_addr[5]=HIBYTE(LOWORD(pd->MacHigh16));

        SMSC_TRACE(DBG_INIT,"Mac Address = %02X:%02X:%02X:%02X:%02X:%02X",
            dev->dev_addr[0],
            dev->dev_addr[1],
            dev->dev_addr[2],
            dev->dev_addr[3],
            dev->dev_addr[4],
            dev->dev_addr[5]);
    return;
}


static ulong pd_set_ids(pPD pd)
{
    if(pd)
      {
        u32 dwIdRev = Lan_GetRegDW(ID_REV);
        u32 dwFpgaRev = Lan_GetRegDW(FPGA_REV);
        ulong name_index;

           
        if(HIWORD(dwIdRev)==LOWORD(dwIdRev)) {
            //this may mean the chip is set for 32 bit
            //  while the bus is reading as 16 bit
            SMSC_WARNING("LAN9420 NOT Identified, dwIdRev==0x%08lX",(ulong)dwIdRev);
            return 1;
        }

        switch(dwIdRev&0xFFFF0000UL) {
            case 0x94200000UL:
                SET_NAME_INDEX(name_index, 0);
            break;
            default:
                SET_NAME_INDEX(name_index, NINDEX_FLT);
            break;
        }
        if(!IS_NAME_INDEX_SET(name_index))
          {
            return 1;
          }
        SMSC_TRACE(DBG_INIT,"  %s identified, dwIdRev==0x%08lX",GET_CHIP_NAME(name_index), (ulong)dwIdRev);
        SMSC_TRACE(DBG_INIT,"  FPGA_REV == 0x%08lX",(ulong)dwFpgaRev);
        SET_PD_REVID(pd,dwIdRev);
        SET_PD_FPGA_REV(pd,dwFpgaRev);
        SET_PD_SIGNATURE(pd);
        return 0;
       }
    return 1;
}

//this function should clear out all leftover structures in PD
//PD itself is be3ing recycled by free_netdev();
static void smsc9420_pd_free(struct net_device *ndev)
{
    pPD pd =  netdev_priv(ndev);
//TxSkBuffers and RxSkBuffers should have been freed in close
//we'll check just in case
    if(pd->TxSkBuffers)
    {
        Tx_buffers_destroy(pd);
    }
    if(pd->RxSkBuffers)
    {
        Rx_buffers_destroy(pd);
    }

    pci_free_consistent(GET_PD_PDEV(pd), 
            sizeof(SMSC9420_DMA_DESCRIPTOR) * pd->rx_ring_size 
            +
            sizeof(SMSC9420_DMA_DESCRIPTOR) * pd->tx_ring_size,
            pd->RxDescRing, 
            pd->rx_dma_addr);
    return;
}

static void __devexit Smsc9420_remove(struct pci_dev *pdev)
{


    struct net_device *ndev;
    pPD pd; 

    SMSC_TRACE(DBG_INIT,"--> %s()", __FUNCTION__);
    SMSC_TRACE(DBG_CLOSE,"--> cleanup_module()");
    ndev = pci_get_drvdata(pdev);
    if(!ndev)
        return;
    pd = netdev_priv(ndev);
    unregister_netdev(ndev);
    smsc9420_pd_free(ndev);
    iounmap((void __iomem *)(GET_PD_LANBASE(pd)-LAN9420_CPSR_ENDIAN_OFFSET));
    free_netdev(ndev);
    pci_release_regions(pdev);
    pci_disable_device(pdev);
    pci_set_drvdata(pdev, NULL);

    SMSC_TRACE(DBG_CLOSE,"<-- cleanup_module()");
}

static int Smsc9420_open(struct net_device *dev)
{
    int result=-ENODEV;

    pPD pd=NULL;
    ulong acquired_irq=0;
    ulong tx_initialized=0;
    ulong rx_initialized=0;

    SMSC_TRACE(DBG_INIT,"-->Smsc9420_open(dev=0x%08lX",(ulong)dev);

#ifndef LINUX_2_6_OR_NEWER
    MOD_INC_USE_COUNT;
#endif

    if(dev==NULL) {
        SMSC_WARNING("dev==NULL");
        result=-EFAULT;
        goto DONE;
    }
    pd=(pPD)netdev_priv(dev);
    if(pd==NULL) {
        SMSC_WARNING("pd==NULL");
        result=-EFAULT;
        goto DONE;
    }

    
    netif_carrier_off(dev);

    if(request_irq(dev->irq, Smsc9420_ISR, 
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22))
                                SA_SHIRQ | SA_INTERRUPT,
#else
                                IRQF_SHARED | IRQF_DISABLED,
#endif

                                "LAN9420_ETH", pd))
        {
                SMSC_WARNING("Unable to use IRQ = %d", dev->irq);
                result=-ENODEV;
                goto DONE;
        }
    Lan_DisableIRQ(pd);
    if(!Lan_Initialize(pd)) {
        SMSC_WARNING("Failed Lan_Initialize");
        result=-ENODEV;
        goto DONE;
    }
    acquired_irq=1;

    //must now test the IRQ connection to the ISR
    Lan_EnableIRQ(pd);
    SMSC_TRACE(DBG_INIT,"Testing ISR using IRQ %d",dev->irq);
    {
        ulong dwTimeOut=1000;
        Lan_SignalSoftwareInterrupt(pd);
        do {
            udelay(10);
            dwTimeOut--;
        } while((dwTimeOut)&&(!(pd->SoftwareInterruptSignal)));
        if(!(pd->SoftwareInterruptSignal)) {
            SMSC_WARNING("ISR failed signaling test");
            result=-ENODEV;
            goto DONE;
        }
    }
    SMSC_TRACE(DBG_INIT,"ISR passed test using IRQ %d",dev->irq);
    Lan_DisableIRQ(pd);


    if(!Tx_Initialize(pd)) {
        SMSC_WARNING("Failed to Initialize the Transmitter");
        result=-ENOMEM;
        goto DONE;
    }
    tx_initialized=1;


    if(!Rx_Initialize(pd)) {
        SMSC_WARNING("Failed to Initialize the Receiver");
        result=-ENOMEM;
        goto DONE;
    }
    rx_initialized=1;

    
    if(!Phy_Initialize(pd,link_mode))
    {
        SMSC_WARNING("Failed to initialize Phy");
        result=-ENODEV;
        goto DONE;
    }

    #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
    //if napi_enable() is called twice BUG_ON happens
    //to prevent this from happening we need to carefully
    //track wether napi's enabled or not.
    //on open most of the intellegence is missing
    //untill better times I just check if NAPI is enabled
    //and call napi_enable only if it is not
    if(test_bit(NAPI_STATE_SCHED, &(&pd->napi)->state))
    {
        napi_enable(&pd->napi);   
    }
    #endif
    

    Lan_SetIntDeas(pd,GET_PD_INTDEAS(pd));

    Tx_hw_start(pd);
    Rx_hw_start(pd);

    //netif_start_queue(dev);
    //stopping the queue because it will wake up when the
    //  link goes up.

   // netif_stop_queue(GET_PD_DEV(pd));
    netif_wake_queue(GET_PD_DEV(pd));

    Lan_SetRegDW(RX_POLL_DEMAND,1UL);//Kick off the RX


    tasklet_enable(&(pd->Smsc9420_PHYTasklet));

    Lan_EnableIRQ(pd);


    result=0;

DONE:
    if(result!=0) {
#ifndef LINUX_2_6_OR_NEWER
        MOD_DEC_USE_COUNT;
#endif
        if(pd!=NULL) {
            if(tx_initialized==1) {
                Lan_DisableIRQ(pd);
                Tx_CleanUp(pd);
            }
            if(rx_initialized==1) {
                Lan_DisableIRQ(pd);
                Rx_CleanUp(pd);
            }
            if(acquired_irq) {
                free_irq(dev->irq, pd);
            }
        }
    }
    SMSC_TRACE(DBG_INIT,"<--Smsc9420_open, result=%d",result);
    return result;
}

static int Smsc9420_stop(struct net_device *dev) 
{
    int result=0;
    pPD pd=NULL;
    SMSC_TRACE(DBG_CLOSE,"-->Smsc9420_stop(dev=0x%08lX)",(ulong)dev);
    if(dev==NULL) {
        SMSC_WARNING("dev==NULL");
        result=-EFAULT;
        goto DONE;
    }
    pd=(pPD)netdev_priv(dev);
    if(pd==NULL) {
        SMSC_WARNING("pd==NULL");
        result=-EFAULT;
        goto DONE;
    }
    

    Lan_DisableIRQ(pd);
    smsc9420_down(dev);
    set_mac_sleep(pd, 1);
    set_phy_sleep(pd, 1);
#ifndef LINUX_2_6_OR_NEWER
    MOD_DEC_USE_COUNT;
#endif
DONE:
    SMSC_TRACE(DBG_CLOSE,"<--Smsc9420_stop, result=%d",result);
    return result;
}


static void smsc9420_down(struct net_device *ndev)
{
    pPD pd=NULL;
    pd=(pPD)netdev_priv(ndev);
    netif_tx_disable(GET_PD_DEV(pd));
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
    napi_disable(&pd->napi);
#endif
    pd->StopLinkPolling=1;
    pd->LanInitialized=1;
    del_timer_sync(&(pd->LinkPollingTimer));
    tasklet_disable(&(pd->Smsc9420_PHYTasklet));
    Tx_CleanUp(pd);
    Rx_CleanUp(pd);  
    free_irq(ndev->irq, pd);
    return;
}


static void Smsc9420_PHYTaskletRoutine(ulong data)
{
    pPD __attribute__ ((unused)) pd=(pPD)data;
    //TODO
    return;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
irqreturn_t Smsc9420_ISR(int Irq,void *dev_id,struct pt_regs *regs)
#else
irqreturn_t Smsc9420_ISR(int Irq,void *dev_id)
#endif
{
    u32 IntCfg=0;
    u32 IntSts=0;

    pPD pd=(pPD)dev_id;

    //accounting on_entry

    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
    regs=regs;//make lint happy
    #endif
    Irq=Irq;//make lint happy

    //sanity checks 
    //we can't continue without the context
    if(unlikely(pd==NULL)) {
        SMSC_WARNING("pd==NULL");
        return IRQ_NONE;
    }
    //we can't continue without the hw
    if(unlikely(!IS_PD_LANBASE_SET(pd))) {
        SMSC_WARNING("LanBase==0");
        return IRQ_NONE;
    }


    IntCfg=Lan_GetRegDW(INT_CFG);

    if((IntCfg&(INT_CFG_IRQ_EN_|INT_CFG_IRQ_INT_))!=
        (INT_CFG_IRQ_EN_|INT_CFG_IRQ_INT_))
    {
        //not our interrupt
        return IRQ_NONE;
    }
    //unexpected config, possible surprise removal
    if(unlikely(IntCfg&SMSC_INT_CFG_RESERVED_BITS )) {
        SMSC_WARNING("In ISR, reserved bits are high.");
        return IRQ_NONE;
    }


    //get interrupt sourceMask
    IntSts=Lan_GetRegDW(INT_STAT);


    if(likely(
              INT_STAT_DMAC_INT_ &
              IntSts
              )) 
    {
        Lan_HandleDMAInt(pd);
    }


    if(unlikely(
                INT_STAT_SW_INT_ & 
                IntSts
                ))
    {
        Lan_HandleSWInt(pd);
    }

    if(unlikely(
                INT_STAT_MBERR_INT_ & 
                IntSts
                )) 
    {
        Lan_HandleMBErrInt(pd);
    }

    if(unlikely(
                INT_STAT_SBERR_INT_ & 
                IntSts
                )) 
    {
        Lan_HandleSBErrInt(pd);
    }

    if(unlikely(
                INT_STAT_GPT_INT_ & 
                IntSts
                ))
    {
        Lan_HandleGPTInt(pd);
    }

    if(unlikely(
                INT_STAT_PHY_INT_ &
                IntSts
                )) 
    {
        Lan_HandlePHYInt(pd);
    }

    if(unlikely(
                INT_STAT_PMT_INT_ &
                IntSts
                )) 
    {
        //we should never see this interrupt
        Lan_HandleWAKEInt(pd);
    }

    if(unlikely(
                INT_STAT_GPIO_INTS_ &
                IntSts
                )) 
    {
        Lan_HandleGPIOInt(pd);
    }

    //since we have just handled an interrupt
    //we have written INT_STS or  INT_CTL or INT_STS or DMAC_STATUS registers
    //to ensure PCI write completion, we must perform a PCI read 
    smsc9420_pci_push(pd);
    return IRQ_HANDLED;
}

static void smsc9420_pci_push(pPD pd)
{
    Lan_GetRegDW(ID_REV);
    return;
}

static void Lan_HandleDMAInt(pPD  pd)
{
        
        u32 DmacSts;
        ulong need_to_wake_tx = 0;
        ulong need_to_schedule_rx = 0;
        ulong isr_repeat_the_loop = GET_PD_ISR_CNT(pd);


        do
        {
            u32 ints_to_clear = 0;
            DmacSts =  Lan_GetRegDW(DMAC_STATUS);
            if(DmacSts&DMAC_STS_TX_)
            {
            //TXDMAC interrupt is never disabled
            //clear
                ints_to_clear |= (DMAC_STS_TX_ | DMAC_STS_NIS_);
                need_to_wake_tx = 1;
            }
            if(DmacSts&DMAC_STS_RX_) 
            {
              //disable RXDMAC interrupt
              Lan_Disable_RXDMACInt_nolock(pd);
              //clear RXDMAC interrupt
              ints_to_clear |= (DMAC_STS_RX_ | DMAC_STS_NIS_);
              need_to_schedule_rx = 1;

            }
            if(DmacSts &(DMAC_STS_TXBU_))
            {
                need_to_wake_tx = 1;
                ints_to_clear |= (DMAC_STS_TXBU_ | DMAC_STS_NIS_);
            }
            //process all the others
            //ERI and ETI must have been cleared already
            //check RWT watchdog TO
            if(DmacSts & DMAC_STS_RWT_)        
            {
                //just clear it and ignore 
//
                ints_to_clear |= DMAC_STS_RWT_;
            }
            if(DmacSts & DMAC_STS_RXPS_)
            {
                //this is tricky
                //since I don't know what to do, let's just clear it
                ints_to_clear |= (DMAC_STS_RXPS_ | DMAC_STS_AIS_);

            }
            if(DmacSts & 
                        DMAC_STS_TXPS_) 
            {
                ints_to_clear |= 
                        (DMAC_STS_TXPS_ | DMAC_STS_AIS_);
            }
            if(DmacSts & DMAC_STS_TXUNF_)
            {
                printk("TXUNF");
                //another bad one
                ints_to_clear |= (DMAC_STS_TXUNF_ |DMAC_STS_AIS_);
            }

            if(ints_to_clear)
            {
                Lan_SetRegDW(DMAC_STATUS, ints_to_clear);
            }

            if(need_to_wake_tx && need_to_schedule_rx)
            {
                //both tx and rx are here, get out
                isr_repeat_the_loop = 0;
            }
        }while(isr_repeat_the_loop--);

        if(need_to_wake_tx)
        {
              netif_wake_queue(GET_PD_DEV(pd));
        }
        if(need_to_schedule_rx)
        {

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
           netif_rx_schedule(GET_PD_DEV(pd));
#else
           netif_rx_schedule(GET_PD_DEV(pd), &pd->napi);
#endif
        }
        return;
}

static void Lan_HandleGPIOInt(pPD pd)
{
//    printk("GPIOINT\n");
    //clear the interrupt
    u32 int_sts = Lan_GetRegDW(INT_STAT);
    int_sts &= (INT_STAT_GPIO_INTS_);
    Lan_SetRegDW(INT_STAT, int_sts);
    return;
}
static void Lan_HandlePHYInt(pPD 
        __attribute__ ((unused))
        pd)
{
    //TODO
//    printk("PHYINT\n");
    return;
}
static void Lan_HandleWAKEInt(pPD pd)
{
    //TODO 
#if 1
    u32 val_PMTCRL, dwWuCsr;
    //PCI handles PME
    //this interrupt is not needed
    //just update the accounting
    Lan_DisableInterrupt(pd,INT_CTL_PMT_INT_EN_);

    dwWuCsr=Lan_GetRegDW(WUCSR);
    val_PMTCRL=Lan_GetRegDW(PMT_CTRL);
            switch(PMT_CTRL_WUPS_MASK_ & val_PMTCRL) 
            {
                case PMT_CTRL_WUPS_PHY_:
                    SMSC_TRACE(DBG_PWR,"PMT_CTRL: Energy Detect activated");
                    //post PHYTasklet command
#if 1
                    Phy_GetRegW(pd,PHY_INT_SRC); //Read to clear interrupt source
#else
                    tasklet_schedule(&(pd->Smsc9420_PHYTasklet));
#endif
                    break;
                case PMT_CTRL_WUPS_MAC_:
                    SMSC_TRACE(DBG_PWR,"PMT_CTRL: Wake on Lan Activated");
                    if(dwWuCsr&WUCSR_WUFR_) {
                        SMSC_TRACE(DBG_PWR,"WUCSR: Wake-Up Frame Received");
                        Lan_SetRegDW(WUCSR,dwWuCsr | WUCSR_WUFR_);//Clear
                    }
                    if(dwWuCsr&WUCSR_MPR_) {
                        SMSC_TRACE(DBG_PWR,"WUCSR: Magic Packet Received");
                        Lan_SetRegDW(WUCSR,dwWuCsr | WUCSR_MPR_);//Clear
                    }
                    break;
                case PMT_CTRL_WUPS_NONE_:
                    break;
                case 0x18UL:SMSC_TRACE(DBG_PWR,"PMT_CTRL: Multiple Events Activated");
                    Lan_SetRegDW(WUCSR,dwWuCsr | WUCSR_MPR_ | WUCSR_WUFR_);//Clear
                    break;
                default:
                    SMSC_WARNING("PMT_INT but No Wake Up Event Activated");
                    break;
            }
#endif
            Lan_SetRegDW(PMT_CTRL,Lan_GetRegDW(PMT_CTRL)|PMT_CTRL_ED_EN_|0x08UL);//clear it

    return;
}
        
                                        
//this function check DMAC RDESC0 status
//updates the counters and gives a go or no-go status for this packet 
static ulong is_rx_packet_good(pPD pd, const ulong ring_index)
{
   u32 desc_status; 
   ulong ignore_crc_error = 0;
   ulong ret = 1;
   //grab DMAC status for this descriptor
   desc_status = (pd->RxDescRing[ring_index].dwStatus);
   //check Rx status
   if(unlikely(RDES0_ERROR_SUMMARY_ &
           desc_status))
   {
       //this is OR of CRC, Collision, Frame to long, Runt frame, Descriptor Error
       pd->stats.rx_errors++;
       if(RDES0_DESCRIPTOR_ERROR_ &
               desc_status)
       {
           //is this the correct counter??
           pd->stats.rx_over_errors++;
       }
       if(RDES0_FRAME_TOO_LONG_ &
               desc_status)
       {
           ignore_crc_error = 1;
           pd->stats.rx_frame_errors++;
       }
       if(RDES0_RUNT_FRAME_ & 
               desc_status)
       {
           //frame too short
           ignore_crc_error = 1;
           pd->stats.rx_frame_errors++;
       }
       if(RDES0_COLLISION_SEEN_ &
               desc_status)
       {
           pd->stats.rx_frame_errors++;
           ignore_crc_error = 1;
       }

       //this bit is invalid if: RuntFrame. Collision,Watchdog TO
       if(!ignore_crc_error 
               && ( RDES0_CRC_ERROR_ &
               desc_status))
       {
           pd->stats.rx_crc_errors++;
       }
       ret = 0;
   }

   if(unlikely(RDES0_LENGTH_ERROR_ &
               desc_status))
   {
        pd->stats.rx_length_errors++;
        ret = 0;

   }
   //each buffer should have LS and FS set
   //chaining in RX is not possible
   if(unlikely ( !
               (RDES0_LAST_DESCRIPTOR_ &
               desc_status ) &&
         (RDES0_FIRST_DESCRIPTOR_ &
          desc_status)
         )
     )
   {
           //the frame spans into multiple buffers
           //should never happen to us 
           //unless there is a jumbo frame
           //which we don't support anyway
           pd->stats.rx_length_errors++;
           ret = 0;
   }
   if(RDES0_MULTICAST_FRAME_ &
           desc_status)
   {
       pd->stats.multicast++;
   }

   return ret;
}


static void discard_rx_packet(pPD 
    __attribute__ ((unused))
        pd, 
        const ulong 
    __attribute__ ((unused))
        index)
{
    //nobody cares about this data
    //we can easily reuse the buffer 
    //in fact we are not event going to unmap it
    //TODO
    //update accounting


    return;
}

//this function attempts to push skb to the stack
//it is important to remember that SKB is pci_mapped 
//and as such we should avoid skb access untill we 
//unmap it inside this function!
//if such data-peek is absolutely neccessary,
//we should use  dma_pci_sync_single_for_cpu()
//on access and possibly in reverse direction
//if we decide not to unmap the buffer and not to deliver it
static void deliver_rx_packet(pPD pd, const ulong index)
{
    struct sk_buff *pSkb = NULL;
    u16 hw_csum = 0;
    struct net_device *ndev = GET_PD_DEV(pd);
    u16 packet_length = (((pd->RxDescRing[index].dwStatus) & 
            (RDES0_FRAME_LENGTH_MASK_)) >> (RDES0_FRAME_LENGTH_SHFT_));

    if(IS_PD_RxCSUM_SET(pd))
    {
        //Adjust length for Rx COE csum
        packet_length -= 2;
    }

    packet_length -=4; //CRC

    if(unlikely(
                packet_length > VLAN_ETH_FRAME_LEN))
    {
        SMSC_WARNING("Rx Packet too long: %d", packet_length);
        return;
    }
    //the code below checks if the frame of VLAN_ETH_FRAME_LEN actually is
    //a VLAN frame.
    //the peeking on  dma_mapped buffer is expensive
    //Furthermore, it appears as if the hw can evaluate the condition
    //and signal us with frame_to_long bit.
    //This needs to be confirmed.
    //In any case, the value of this evaluation does not outweight 
    //the cost of pci_dma_sync_single for the entire VLAN_ETH_FRAME_LEN
#if 0
    //validate packet length
    switch(packet_length)
    {
        case VLAN_ETH_FRAME_LEN:
        //we need to peek inside the buffer
        //we should first take the ownership of this buffer
        //this is a bad thing to do....
            pci_dma_sync_single_for_cpu(GET_PD_PDEV(pd), 
                        pd->RxSkBuffers[index].mapping_RxSkBuffers,
                        NET_IP_ALIGN + packet_length + 2,
                        PCI_DMA_FROMDEVICE);
            //check if this is indeed a VLAN frame
            if(unlikely(get_unaligned_le16((void *)
                            ((pd->RxSkBuffers[index].RxSkb->data) +
                            NET_IP_ALIGN + 12))  !=
                        ntohs(ETH_P_8021Q)))
            {
                pd->stats.rx_errors++;
                pd->stats.rx_frame_errors++;
                //return the buffer back to dma
                pci_dma_sync_single_for_device(GET_PD_PDEV(pd),
                                pd->RxSkBuffers[index].mapping_RxSkBuffers,
                                NET_IP_ALIGN + packet_length + 2,
                                PCI_DMA_TODEVICE);
                SMSC_WARNING("Rx Packet length error: %d",packet_length);
                return;
            }
            break;
        case ETH_FRAME_LEN:
            break;
        default:

            break;
    }
#endif

    pd->stats.rx_packets++;
    pd->stats.rx_bytes += packet_length;

    //now we have 2 options:
    //1. allocate a new skb, copy the data and avoid sync 
    //2. purge cache, deliver skb and allocate a new big one
    //There is a possiblity that option 1 is beneficial for small packets 
    //on certain CPUs.
    //One can attempt to select which option to use based on data size;
    //Since actual measurements of this optimisation is not known +
    //x86 apparently wouldn't benefit from this at all,
    //I decided to omit option 1 all together, untill more experimental data
    //is collected

    pci_unmap_single(GET_PD_PDEV(pd),
            pd->RxSkBuffers[index].mapping_RxSkBuffers,
            PKT_BUF_SZ,
            PCI_DMA_FROMDEVICE);
    pd->RxSkBuffers[index].mapping_RxSkBuffers = 0;

    pSkb = pd->RxSkBuffers[index].RxSkb;
    pd->RxSkBuffers[index].RxSkb = NULL;;


    if(likely(IS_PD_RxCSUM_SET(pd)))
    {
        //Grab the RX COE csum, which is the last 2 bytes (after CRC)
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22))
         hw_csum =
         get_unaligned((u16 *)(pSkb->tail + 
                NET_IP_ALIGN + 
                packet_length + 4));
#else
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26))
        hw_csum = get_unaligned((u16*)(skb_tail_pointer(pSkb) + 
                NET_IP_ALIGN + 
                packet_length + 4));
#else   //get_unaligned_le16 is available
        hw_csum = get_unaligned_le16((skb_tail_pointer(pSkb) + 
                NET_IP_ALIGN + 
                packet_length + 4));
#endif  //2.6.26
#endif //2.6.22
        //the 4 in the above function came from packet_length -= 4 above:)
    }

    skb_reserve(pSkb, NET_IP_ALIGN);
    skb_put(pSkb, packet_length);

    if(likely(IS_PD_RxCSUM_SET(pd)))
    {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
        pSkb->ip_summed = CHECKSUM_HW;
#else
        pSkb->ip_summed = CHECKSUM_COMPLETE;
#endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26))

        put_unaligned(cpu_to_le16(hw_csum), (u16*) &(pSkb->csum));
#else
        put_unaligned_le16(cpu_to_le16(hw_csum), (u16*) &(pSkb->csum));
#endif //2.6.26
    }

    pSkb->protocol = eth_type_trans(pSkb, GET_PD_DEV(pd));

    netif_receive_skb(pSkb);
    ndev->last_rx = jiffies;
    return;
}

static ulong rx_proc_dmac(pPD pd, ulong limit)
{
    ulong cnt = limit;
    ulong drop_frame_cnt= 0;
    ulong index;
    do
    {
        if(unlikely(!Rx_GetFreeDescriptorCount(pd)))
        {
            //there is no more descriptors we can process
            break;
        }
        index = pd->RxHeadIndex;
        //check if we own this dma descriptor
        if((RDES0_OWN_) &
                pd->RxDescRing[index].dwStatus)
        {
            break;
        }
        if(!is_rx_packet_good(pd, index))
        {
            //skip this one
            discard_rx_packet(pd, index);
            rx_reuse_old_buffer(pd, index);
        }
        else
        {
            deliver_rx_packet(pd,index);
            rx_set_new_buffer(pd,index);
        }
        Rx_IncrementHeadIndex(pd);
    }while(--limit);

    drop_frame_cnt = Lan_GetRegDW(MISS_FRAME_CNTR);
    //27:17 contains MIL_FIFL_FULL count 
    //15:0  countains RX_BUFF_UNAV count 
    pd->stats.rx_dropped+=
        (drop_frame_cnt& 0xFFFF) +
        ((drop_frame_cnt >> 17) & 0x3FF);

    return (cnt-limit);
}

static void rx_reuse_old_buffer(pPD pd, const ulong index)
{
    if(unlikely(!(pd->RxSkBuffers[index].RxSkb)))
    {
        //can't reuse this buffer, before giving up. let's 
        //check the mapping 
        if(unlikely(pd->RxSkBuffers[index].mapping_RxSkBuffers))
        {
            pci_unmap_single(GET_PD_PDEV(pd), 
                    pd->RxSkBuffers[index].mapping_RxSkBuffers,
                    PKT_BUF_SZ, PCI_DMA_FROMDEVICE);
            pd->RxSkBuffers[index].mapping_RxSkBuffers = 0;

        }
        rx_set_new_buffer(pd, index);
        return;
    }
    else 
    {
        //there is an skb 
        //check there is a mapping 
        if(likely(pd->RxSkBuffers[index].mapping_RxSkBuffers))
        {
            //do rearming here
            pd->RxDescRing[index].dwBuffer1 =
                (u32)((u32)pd->RxSkBuffers[index].mapping_RxSkBuffers + NET_IP_ALIGN);
            pd->RxDescRing[index].dwStatus = (RDES0_OWN_);
            Rx_IncrementTailIndex(pd);
        }
        else
        {
            //there is skb, but no mapping;
            dev_kfree_skb_any(pd->RxSkBuffers[index].RxSkb);
            pd->RxSkBuffers[index].RxSkb = NULL;
            rx_set_new_buffer(pd, index);
        }
    }
    return;
}
//this function allocates new rx skb,
//creates pci_mapping, updates RxDMA descriptor
//however we do not release the ownership  of the
//descriptor just yet, nor we poll here
//
static void rx_set_new_buffer(pPD pd, const ulong index)
{
    struct sk_buff *pSkb = NULL;
    dma_addr_t mapping;
    if(unlikely(!
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
                (pSkb = dev_alloc_skb(PKT_BUF_SZ))
#else
                (pSkb = netdev_alloc_skb(GET_PD_DEV(pd),  PKT_BUF_SZ))
#endif
               )
        )
    {
        SMSC_WARNING("Failed to allocate new skb!");
        return;
    }
    pSkb->dev = GET_PD_DEV(pd);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22))
    mapping = pci_map_single(GET_PD_PDEV(pd),
            (ulong *)pSkb->tail, PKT_BUF_SZ,
            PCI_DMA_FROMDEVICE);
#else
    mapping = pci_map_single(GET_PD_PDEV(pd),
            (ulong *)skb_tail_pointer(pSkb),
            PKT_BUF_SZ,
            PCI_DMA_FROMDEVICE);
#endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27))
    if(pci_dma_mapping_error(mapping))
#else
    if(pci_dma_mapping_error(GET_PD_PDEV(pd), mapping))
#endif
    {
        dev_kfree_skb_any(pSkb);
        SMSC_WARNING("pci_map_single failed!");
        return;
    }

    pd->RxSkBuffers[index].RxSkb = pSkb;
    pd->RxSkBuffers[index].mapping_RxSkBuffers = mapping;
    pd->RxDescRing[index].dwBuffer1 = 
        (u32)((u32)mapping + NET_IP_ALIGN);
    pd->RxDescRing[index].dwStatus = (RDES0_OWN_);
    Rx_IncrementTailIndex(pd);
    return;
}


static void smsc9420_rx_clean(pPD pd, ulong *work_done, ulong budget)
{

    //process the data
    *work_done = rx_proc_dmac(pd, budget);
    //Kick RXDMA
    Lan_SetRegDW(RX_POLL_DEMAND, (1UL));
    smsc9420_pci_push(pd);
    return;
}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
static int Smsc9420_rx_poll(struct net_device *dev,int * budget)
#else
static int Smsc9420_rx_poll(struct napi_struct *napi, int budget)
#endif
{

    ulong work_done = 0;
    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
    int result=0;
    pPD pd;
    pd=((pPD)netdev_priv(dev));
    #else
	pPD pd;
	struct net_device *dev;

	pd = container_of(napi, PRIVATE_DATA, napi);
	dev = GET_PD_DEV(pd);
    #endif


    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
    SET_PD_RX_WLIMIT(pd, dev->quota);
    if(GET_PD_RX_WLIMIT(pd) >
            (ulong)(*budget)) 
    {
        SET_PD_RX_WLIMIT(pd,(ulong)(*budget));
    }
    #else
    SET_PD_RX_WLIMIT(pd,budget);
    #endif

    smsc9420_rx_clean(pd, &work_done, GET_PD_RX_WLIMIT(pd));

    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
    dev->quota-=work_done;
    (*budget)-=work_done;
    #endif

    if(work_done < GET_PD_RX_WLIMIT(pd))
    {
        #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
        netif_rx_complete(dev);
        #else
        netif_rx_complete(dev, &pd->napi);
        #endif
        if(IS_CLR_TIMER_ON_RX(pd)) 
        {
            Lan_ClearDeassertionTimer(pd);
        }
        Lan_Enable_RXDMACInt_nolock(pd);
    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
    } else {
        result=1;
    #endif
    }
    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
    return result;
    #else
    return work_done;
    #endif
}


//this function executes in softirq context
//at present time Tx path works like this:
//at hard_start_xmit we:
//1.recycle all the previously processed skbs
//2.place new skb into Tx DMA ring
//3.if we are taking the last descriptor we 
//4.stop the queue
//5.request an interrupt upon closing *this* descriptor
//this interrupt will wake the tx queue.
//if this is not the last descriptor no work is required
//BENEFITS: because linux does not care about the actual tx status of the skb
//we do not need an ISR or any flavor of deffered processing so long as
//we do not run out of descriptors. It sounds as if ISR is not telling us anything we
//would care to know. Interrupt is usefull, however, to tell us
//that we have new descriptos available.(once we ran out of them).
//Another positive consideration is that: hard_start_xmit runs in the context of
//softirq. It can not be preempted by anything other than ISR. IF we agree that ISR
//does not share data with hard_start_xmit() there is absolutely 
//no synchronization required.
//PROBLEMS: we will have to use up all the descriptors before stoping the queue. 
//at this point we will request an interrupt for the current buffer.
//this would mean that we have to wait untill *all* the decriptors are processed
//before ISR will wake the queue again.
//ideally, we would like to interrupt on dma descriptors half used.
//that is not easy now to coordinate queue wake/stop and ISR.
//THIS needs to be architected later!@!!!
//in particular: needs_tx_queue_stop(), needs_tx_dmac_int() functions 
//needs to be redesigned.
//-----------------
//SOLUTIONS:??? 
//Since wake_queue within ISR is  does not hurt already running queue
//we can safely interrupt on half the ring.
//one extra interrupt is a small penalty to pay, while benefits of 
//having no enforced gaps in transmission are clear.
//it is not clear how efficient half ring interrupt is. but it definitely
//helps.


//return NETDEV_TX_OK  || NETDEV_TX_BUSY 
 
static int Smsc9420_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{

    int result=
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9))
        0;
#else
        NETDEV_TX_OK;
#endif
    pPD pd=NULL;
    pd=(pPD)netdev_priv(dev);
    //clean Tx
    recycle_tx_dma(pd);
    //start new Tx
    if(!put_new_tx_dma(pd, skb))
    {
        //we have failed delivery
        //return the skb
        result = 
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9))
            1;
#else
            NETDEV_TX_BUSY;
#endif

    }
    dev->trans_start = jiffies;

    return result;
}

//this funcion decides if we need TX completion interrupt
static ulong needs_tx_dmac_int(pPD pd)
{
    int delta = TX_GET_HEAD(pd) - TX_GET_TAIL(pd);
    if(delta < 0)
    {
        delta = 0 - delta;
    }

    if((ulong)delta == (pd->tx_ring_size / 2))
    {
        return 1;
    }
    //and when
    //we are about to take the last descriptor 
    return ((TX_NEW_HEAD(pd) == TX_GET_TAIL(pd)) ? 1 : 0);
}
static ulong needs_tx_queue_stop(pPD pd)
{
    return ((TX_NEW_HEAD(pd) == TX_GET_TAIL(pd)) ? 1 : 0);
}
static ulong put_new_tx_dma(pPD pd, struct sk_buff* skb)
{
    dma_addr_t mapping;
    ulong index;
    u32 tmp_desc1;
    if(unlikely(skb->len < ETH_ZLEN))
    {
        //TODO
        //just record the situation 
        //hw takes care of the short frames 
    }
    if(TX_IS_FULL(pd))
    {
//        SMSC_WARNING("Ring is full! Dropping the packet");
        return 0;
    }

    index = TX_GET_HEAD(pd);
    if(unlikely(
                TDES0_OWN_ & 
                (pd->TxDescRing[index].dwStatus)))
    {
        SMSC_WARNING("%s():%d:TxDMAC OWN bit race!",
                __FUNCTION__, __LINE__);
        return 0;
    }

    if(unlikely(needs_tx_queue_stop(pd)))
    {
        netif_stop_queue(GET_PD_DEV(pd));
    }
    pd->TxSkBuffers[index].TxSkb = skb;
    pd->TxSkBuffers[index].len=skb->len;
    mapping = pci_map_single(GET_PD_PDEV(pd), 
            skb->data, skb->len, PCI_DMA_TODEVICE);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27))
    if(pci_dma_mapping_error(mapping))
#else
    if(pci_dma_mapping_error(GET_PD_PDEV(pd), mapping))
#endif
    {
        //do not free this skb
//        dev_kfree_skb_any(skb);
        pd->TxSkBuffers[index].TxSkb = NULL;
        pd->TxSkBuffers[index].len = 0;
        //update the stats 
        SMSC_WARNING("pci_map_single failed!");
        return 0;
    }
    pd->TxSkBuffers[index].mapping_TxSkBuffers = mapping;

    pd->TxDescRing[index].dwBuffer1 = (u32)(mapping);
    tmp_desc1 = (TDES1_LS_ | (((u32)skb->len) & 0x7FF));
    if(unlikely(needs_tx_dmac_int(pd)))
    {
        tmp_desc1 |= TDES1_IC_;
    }
    //check if we are at the last descriptor and need to set EOR
    if(unlikely(index == (pd->tx_ring_size - 1)))
    {
       tmp_desc1 |= TDES1_TER_;
    }
    pd->TxDescRing[index].dwLength = (tmp_desc1);

    TX_INC_HEAD(pd);
    //assign ownership to DMAC

    wmb();
    pd->TxDescRing[index].dwStatus = (u32)(TDES0_OWN_);
    wmb();
    //kick the DMA 
    Lan_SetRegDW(TX_POLL_DEMAND,0xFFFFFFFFUL);
    smsc9420_pci_push(pd);

    return 1;
}
static void tx_eval_status(pPD pd, const ulong status, const ulong index)
{
        if(unlikely(TDES0_ERROR_SUMMARY_ & status))
        {
            //DMA reports an error
            pd->stats.tx_errors++;
            if((TDES0_EXCESSIVE_DEFERRAL_ |
                        TDES0_EXCESSIVE_COLLISIONS_) &
                    status)
            {
                pd->stats.tx_aborted_errors++;
            }
            if((TDES0_LOSS_OF_CARRIER_ |
                        TDES0_NO_CARRIER_))
            {
                pd->stats.tx_carrier_errors++;
            }
        }
        else
        {
            pd->stats.tx_packets++;
            pd->stats.tx_bytes += 
                ((pd->TxDescRing[index].dwLength)&(0x7FFUL));
        }

        if(unlikely((TDES0_EXCESSIVE_COLLISIONS_)&
                    status))
        {
            pd->stats.collisions+=16;
        } else {
            pd->stats.collisions+=
                (status&TDES0_COLLISION_COUNT_MASK_)>>
                (TDES0_COLLISION_COUNT_SHFT_);
        }
        if(unlikely(TDES0_HEARTBEAT_FAIL_
                    & status))
        {
            pd->stats.tx_heartbeat_errors++;
        }
    return;
}

static void recycle_tx_dma(pPD pd)
{
    ulong index;
    ulong status;
    while(likely(!(TX_IS_EMPTY(pd))))
    {

        index = TX_GET_TAIL(pd);

        status = (pd->TxDescRing[index].dwStatus);

        if(unlikely(TDES0_OWN_ & status))
        {
            //it's ok to hit this!
            //we are attempting to recycle  all the descriptors
            //here, some of them are very likely to be still owned 
            //by TXDMA. 
            //DMA owns this descriptor
            break;
        }
        //DMA is done, lets check the status
        tx_eval_status(pd, status, index);

        //now recycle the buffers
        if(likely(pd->TxSkBuffers[index].TxSkb))
        {
            if(unlikely(!pd->TxSkBuffers[index].mapping_TxSkBuffers))
            {
                //there is an SKB,
                //but there is no mapping for it
                //how so?
                //throw it away, complain, continue to the next
                dev_kfree_skb_any(pd->TxSkBuffers[index].TxSkb); 
                SMSC_WARNING("Unmapped SKB in Tx Ring!");
                pd->TxSkBuffers[index].TxSkb=NULL;
                pd->TxSkBuffers[index].len = 0;

                //before we get out we need to advance the index
                //otherwise we'll be spinning
                TX_INC_TAIL(pd);

                continue;
            }
            //ok there is a mapping
            //unmap
            pci_unmap_single(GET_PD_PDEV(pd),
                    pd->TxSkBuffers[index].mapping_TxSkBuffers,
                    pd->TxSkBuffers[index].len,
                    PCI_DMA_TODEVICE);
            //discard skb
            dev_kfree_skb_any(pd->TxSkBuffers[index].TxSkb); 
        }
        else
        {
            //ok there is no skb.
            //how is this possible??
            //we should, however, make sure there is no 
            //outstanding mapping here
            if(pd->TxSkBuffers[index].mapping_TxSkBuffers)
            {
                //sure enough, there is mapping without an skb
                //how is this possible??
                printk("%s():%d:No skb in tx recycle\n", __FUNCTION__, __LINE__);
                pci_unmap_single(GET_PD_PDEV(pd),
                        pd->TxSkBuffers[index].mapping_TxSkBuffers,
                        pd->TxSkBuffers[index].len,
                        PCI_DMA_TODEVICE);
            }
            else
            {
                //no skb, no mapping: unusual but consistent
                printk("%s():%d:No skb in tx recycle\n", __FUNCTION__, __LINE__);
            }

        }
        pd->TxSkBuffers[index].TxSkb=NULL;
        pd->TxSkBuffers[index].mapping_TxSkBuffers=0;
        pd->TxSkBuffers[index].len=0;
        //record this situation with  tx descriptor
        pd->TxDescRing[index].dwBuffer1=(u32)(0UL);

        TX_INC_TAIL(pd);


    }//while
    return;
}

static struct net_device_stats * Smsc9420_get_stats(struct net_device * const dev)
{
    pPD pd=NULL;
    u32 counter;
    if(dev==NULL) {
        SMSC_WARNING("dev==NULL");
        return NULL;
    }
    if(dev->priv==NULL) {
        SMSC_WARNING("dev->priv==NULL");
        return NULL;
    }

    pd=(pPD)netdev_priv(dev);
        counter=Lan_GetRegDW(MISS_FRAME_CNTR);
        pd->stats.rx_dropped+=
            (counter&0x0000FFFFUL)+
            ((counter>>17)&0x000003FF);
    return &(pd->stats);
}


static void Smsc9420_set_multicast_list(struct net_device *dev)
{
    pPD pd=NULL;

    pd=netdev_priv(dev); 
   
    pd->RxFoxFilter=0;
    pd->RxFoxHashLo=0;
    pd->RxFoxHashHi=0;
    
    if(((GET_PD_DEV(pd))->flags&IFF_PROMISC)) {
        pd->RxFoxFilter|=RX_FILTER_PROMISCUOUS;
        goto SET_MULTICAST;
    }
    else
    {
        pd->RxFoxFilter &= (~RX_FILTER_PROMISCUOUS);
    }
    if((GET_PD_DEV(pd))->flags & IFF_ALLMULTI) {
        pd->RxFoxFilter|=RX_FILTER_ALL_MULTICAST;
        goto SET_MULTICAST;
    }
    else
    {
        pd->RxFoxFilter &= (~RX_FILTER_ALL_MULTICAST);

    }
    
    if((GET_PD_DEV(pd))->mc_count>0) {
        u32 dwHashH=0;
        u32 dwHashL=0;
        u32 dwCount=0;
        struct dev_mc_list *mc_list=pd->dev->mc_list;
        while(mc_list!=NULL) {
            dwCount++;
            if(likely(ETH_ALEN == 
                            (mc_list->dmi_addrlen))) {

                u32 dwMask=0x01UL;
                u32 dwBitNum=Rx_Hash(mc_list->dmi_addr);
                dwMask<<=(dwBitNum&0x1FUL);
                if(dwBitNum&0x20UL) {
                    dwHashH|=dwMask;
                } else {
                    dwHashL|=dwMask;
                }
            } else {
                SMSC_WARNING("dmi_addrlen!= ETH_ALEN");
            }
            mc_list=mc_list->next;
        }
        if(dwCount!=(u32)(pd->dev->mc_count)) {
            SMSC_WARNING("dwCount!=dev->mc_count");
        }
        pd->RxFoxHashLo=dwHashL;
        pd->RxFoxHashHi=dwHashH;
    
    }
SET_MULTICAST:
    Rx_SetMulticastList(pd);
    return;
}


static int Smsc9420_ethtool_ioctl(pPD pd, void __user * userAddr)
{
    int result=-EFAULT;
    u32 ethcmd=0;
    if(copy_from_user(&ethcmd,userAddr,sizeof(ethcmd)))
    {
        result=-EFAULT;
        goto DONE;
    }
    switch(ethcmd) {
    case ETHTOOL_GSET:/* Get settings. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GSET");
        {
            struct ethtool_cmd settings={ETHTOOL_GSET};
            settings.supported=
                SUPPORTED_10baseT_Half |
                SUPPORTED_10baseT_Full |
                SUPPORTED_100baseT_Half |
                SUPPORTED_100baseT_Full |
                SUPPORTED_Autoneg |
                SUPPORTED_MII;
            settings.advertising=ADVERTISED_MII;
            if(pd->dwLinkSettings & LINK_SPEED_10HD)
                settings.advertising|=ADVERTISED_10baseT_Half;
            if(pd->dwLinkSettings & LINK_SPEED_10FD)
                settings.advertising|=ADVERTISED_10baseT_Full;
            if(pd->dwLinkSettings & LINK_SPEED_100HD)
                settings.advertising|=ADVERTISED_100baseT_Half;
            if(pd->dwLinkSettings & LINK_SPEED_100FD)
                settings.advertising|=ADVERTISED_100baseT_Full;
            if(pd->dwLinkSettings & LINK_AUTO_NEGOTIATE) {
                settings.advertising|=ADVERTISED_Autoneg;
                settings.autoneg=AUTONEG_ENABLE;
            } else settings.autoneg=AUTONEG_DISABLE;
            if(pd->dwLinkSpeed & (LINK_SPEED_100HD|LINK_SPEED_100FD))
                settings.speed=SPEED_100;
            else settings.speed=SPEED_10;
            if(pd->dwLinkSpeed & (LINK_SPEED_10FD|LINK_SPEED_100FD))
                settings.duplex=DUPLEX_FULL;
            else settings.duplex=DUPLEX_HALF;
            settings.port=PORT_MII;
            settings.phy_address=(u8)pd->dwPhyAddress;
            settings.transceiver=XCVR_INTERNAL;
            settings.maxtxpkt=0;
            settings.maxrxpkt=0;
            if(copy_to_user(userAddr,&settings,sizeof(settings))) {
                result=-EFAULT;
            } else {
                result=0;
            }
        }
        break;
    case ETHTOOL_SSET:/* Set settings. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_SSET");
        {
            struct ethtool_cmd settings;
            u16 speed=0;
            u8 duplex=0;
            u8 autoneg=0;
            if(copy_from_user(&settings,userAddr,sizeof(settings))) {
                result=-EFAULT;
                goto DONE;
            }
            if(pd->dwLinkSettings&LINK_AUTO_NEGOTIATE) {
                autoneg=AUTONEG_ENABLE;
            } else {
                autoneg=AUTONEG_DISABLE;
            }
            if(pd->dwLinkSpeed&(LINK_SPEED_100HD|LINK_SPEED_100FD))
            {
                speed=SPEED_100;
            } else {
                speed=SPEED_10;
            }
            if(pd->dwLinkSpeed&(LINK_SPEED_10FD|LINK_SPEED_100FD))
            {
                duplex=DUPLEX_FULL;
            } else {
                duplex=DUPLEX_HALF;
            }
            if((settings.speed!=100)&&(settings.speed!=10)) {
                result=-EOPNOTSUPP;
                goto DONE;
            }
            if((settings.duplex!=DUPLEX_FULL)&&(settings.duplex!=DUPLEX_HALF)) {
                result=-EOPNOTSUPP;
                goto DONE;
            }
            if((settings.autoneg!=AUTONEG_ENABLE)&&(settings.autoneg!=AUTONEG_DISABLE)) {
                result=-EOPNOTSUPP;
                goto DONE;
            }
            if((settings.autoneg!=autoneg)||
                (settings.speed!=speed)||
                (settings.duplex!=duplex))
            {
                if(settings.autoneg==AUTONEG_ENABLE) {
                    u16 wAdvertisement=Phy_GetRegW(pd,PHY_ANEG_ADV);
                    wAdvertisement&=(~PHY_ANEG_ADV_SPEED_);
                    if(settings.speed==SPEED_100) {
                        if(settings.duplex==DUPLEX_FULL) {
                            wAdvertisement|=PHY_ANEG_ADV_100F_;
                        } else {
                            wAdvertisement|=PHY_ANEG_ADV_100H_;
                        }
                    } else {
                        if(settings.duplex==DUPLEX_FULL) {
                            wAdvertisement|=PHY_ANEG_ADV_10F_;
                        } else {
                            wAdvertisement|=PHY_ANEG_ADV_10H_;
                        }
                    }
                    Phy_SetRegW(pd,PHY_ANEG_ADV,wAdvertisement);
                    Phy_SetRegW(pd,PHY_BCR,
                        (PHY_BCR_ANEG_EN_|PHY_BCR_RST_ANEG_));
                } else {
                    u16 wBcr=Phy_GetRegW(pd,PHY_BCR);
                    if(settings.speed==SPEED_100) {
                        wBcr|=PHY_BCR_SPEED_SELECT_;
                    } else {
                        wBcr&=(~PHY_BCR_SPEED_SELECT_);
                    }
                    if(settings.duplex==DUPLEX_FULL) {
                        wBcr|=PHY_BCR_DUPLEX_MODE_;
                    } else {
                        wBcr&=(~PHY_BCR_DUPLEX_MODE_);
                    }
                    Phy_SetRegW(pd,PHY_BCR,wBcr);
                }
            }
            result=0;
        }
        break;
    case ETHTOOL_GDRVINFO:/* Get driver info. */
//        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GDRVINFO");
        {
            struct ethtool_drvinfo info = {ETHTOOL_GDRVINFO};
            strcpy(info.driver,"Smsc9420_eth");
            memset(&info.version,0,sizeof(info.version));
            strcpy(info.version, DRV_VERSION);
            memset(&info.fw_version,0,sizeof(info.fw_version));
            sprintf(info.fw_version,"%lu",(GET_PD_REVID(pd))&0xFFFFUL);
            memset(&info.bus_info,0,sizeof(info.bus_info));
            memset(&info.reserved1,0,sizeof(info.reserved1));
            memset(&info.reserved2,0,sizeof(info.reserved2));
            info.n_stats=0;
            info.testinfo_len=0;
            info.eedump_len=0;
            info.regdump_len=0;
            if(copy_to_user(userAddr,&info,sizeof(info))) {
                result=-EFAULT;
            } else {
                result=0;
            }
        }
        break;
    case ETHTOOL_GREGS:/* Get NIC registers. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GREGS");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GWOL:/* Get wake-on-lan options. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GWOL");
        {
            struct ethtool_wolinfo wol_info={ETHTOOL_GWOL};
            //wol_info.supported=1;
            wol_info.supported=( WAKE_UCAST | WAKE_BCAST | 
                                 WAKE_MCAST | WAKE_ARP | 
                                 WAKE_MAGIC |
#ifdef USE_WAKE_PHY_ON_PCI
                                 WAKE_PHY   |
#endif
                                 0);
            wol_info.wolopts= GET_PD_WOLOPTS(pd); 
            memset(&wol_info.sopass,0,sizeof(wol_info.sopass));
            if(copy_to_user(userAddr,&wol_info,sizeof(wol_info))) {
                result=-EFAULT;
            } else {
                result=0;
            }
        }
        break;
    case ETHTOOL_SWOL:/* Set wake-on-lan options. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_SWOL");
        {
            struct ethtool_wolinfo wol_info;        
            if(copy_from_user(&wol_info,userAddr,sizeof(wol_info)))
            {
                result=-EFAULT;
            } else {
                SMSC_TRACE(DBG_IOCTL,"WOL OPTS = 0x%x", wol_info.wolopts);          
                GET_PD_WOLOPTS(pd)= wol_info.wolopts;
                result=0;
            }
        }
        break;
    case ETHTOOL_GMSGLVL:/* Get driver message level */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GMSGLVL");
        {
            struct ethtool_value msgLevel={ETHTOOL_GMSGLVL};
            msgLevel.data=GET_PD_DEBUG_MODE(pd);
            if(copy_to_user(userAddr, &msgLevel,sizeof(msgLevel))) {
                result=-EFAULT;
            } else {
                result=0;
            }
        }
        break;
    case ETHTOOL_SMSGLVL:/* Set driver msg level. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_SMSGLVL");
        {
            struct ethtool_value msgLevel;
            if(copy_from_user(&msgLevel,userAddr,sizeof(msgLevel)))
            {
                result=-EFAULT;
            } else {
                SET_PD_DEBUG_MODE(pd,msgLevel.data);
                result=0;
            }
        }
        break;
    case ETHTOOL_NWAY_RST:/* Restart autonegotiation. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_NWAY_RST");
        {
            Phy_SetRegW(pd,PHY_BCR,
                (PHY_BCR_ANEG_EN_|PHY_BCR_RST_ANEG_));
            result=0;
        }
        break;
    case ETHTOOL_GLINK:/* Get link status (ethtool_value) */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GLINK");
        {
            struct ethtool_value linkStatus={ETHTOOL_GLINK};
            if(pd->dwLinkSpeed!=LINK_OFF) {
                linkStatus.data=1;
            } else {
                linkStatus.data=0;
            }
            if(copy_to_user(userAddr,&linkStatus,sizeof(linkStatus)))
            {
                result=-EFAULT;
            } else {
                result=0;
            }
        }
        break;
    case ETHTOOL_GEEPROM:/* Get EEPROM data */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GEEPROM");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_SEEPROM:/* Set EEPROM data. */     
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_SEEPROM");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GCOALESCE:/* Get coalesce config */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GCOALESCE");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_SCOALESCE:/* Set coalesce config. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_SCOALESCE");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GRINGPARAM:/* Get ring parameters */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GRINGPARAM");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_SRINGPARAM:/* Set ring parameters. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_SRINGPARAM");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GPAUSEPARAM:/* Get pause parameters */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GPAUSEPARAM");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_SPAUSEPARAM:/* Set pause parameters. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_SPAUSEPARAM");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GRXCSUM:/* Get RX hw csum enable (ethtool_value) */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GRXCSUM");
        {
            struct ethtool_value rxCsum={ETHTOOL_GRXCSUM};
            rxCsum.data=0;
            if(copy_to_user(userAddr,&rxCsum,sizeof(rxCsum))) {
                result=-EFAULT;
            } else {
                result=0;
            }
        }
        break;
    case ETHTOOL_SRXCSUM:/* Set RX hw csum enable (ethtool_value) */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_SRXCSUM");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GTXCSUM:/* Get TX hw csum enable (ethtool_value) */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GTXCSUM");
        {
            struct ethtool_value txCsum={ETHTOOL_GTXCSUM};
            txCsum.data=0;
            if(copy_to_user(userAddr,&txCsum,sizeof(txCsum))) {
                result=-EFAULT;
            } else {
                result=0;
            }
        }
        break;
    case ETHTOOL_STXCSUM:/* Set TX hw csum enable (ethtool_value) */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_STXCSUM");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GSG:/* Get scatter-gather enable (ethtool_value) */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GSG");
        {
            struct ethtool_value sg={ETHTOOL_GSG};
            sg.data=0;
            if(copy_to_user(userAddr,&sg,sizeof(sg))) {
                result=-EFAULT;
            } else {
                result=0;
            }
        }
        break;
    case ETHTOOL_SSG:/* Set scatter-gather enable (ethtool_value). */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_SSG");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_TEST:/* execute NIC self-test. */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_TEST");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GSTRINGS:/* get specified string set */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GSTRINGS");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_PHYS_ID:/* identify the NIC */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_PHYS_ID");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GSTATS:/* get NIC-specific statistics */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GSTATS");
        result=-EOPNOTSUPP;
        break;
    case ETHTOOL_GTSO:/* Get TSO enable (ethtool_value) */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_GTSO");
        {
            struct ethtool_value tso={ETHTOOL_GTSO};
            tso.data=0;
            if(copy_to_user(userAddr,&tso,sizeof(tso))) {
                result=-EFAULT;
            } else {
                result=0;
            }
        }
        break;
    case ETHTOOL_STSO:/* Set TSO enable (ethtool_value) */
        SMSC_TRACE(DBG_IOCTL,"ETHTOOL_STS0");
        result=-EOPNOTSUPP;
        break;

    default:
        SMSC_WARNING("unknown ethcmd=0x%08X",ethcmd);
        result=-EOPNOTSUPP;
        break;
    }
DONE:
    return result;
}

static int Smsc9420_do_ioctl(
    struct net_device *dev,
    struct ifreq *ifr,
    int cmd)
{
    int result=0;
    pPD pd=NULL;
    void __user *userAddr=NULL;

    
    if(dev==NULL) {
        SMSC_WARNING("dev==NULL");
        result=-EFAULT;
        goto DONE;
    }
    if(dev->priv==NULL) {
        SMSC_WARNING("dev->priv==NULL");
        result=-EFAULT;
        goto DONE;
    }
    pd=((pPD)netdev_priv(dev));
    if(ifr==NULL) {
        SMSC_WARNING("ifr==NULL");
        result=-EFAULT;
        goto DONE;
    }
    userAddr=ifr->ifr_data;

    switch(cmd) {
    case SIOCGMIIPHY:
    case SIOCDEVPRIVATE:
        SMSC_TRACE(DBG_IOCTL,"SIOCGMIIPHY");
        if(pd->LanInitialized) {
            struct mii_ioctl_data *miiData=
                (struct mii_ioctl_data *)&(ifr->ifr_data);
            miiData->phy_id=1;
        };break;
    case SIOCGMIIREG:
    case SIOCDEVPRIVATE+1:
        SMSC_TRACE(DBG_IOCTL,"SIOCGMIIREG");
        if(pd->LanInitialized) {
            struct mii_ioctl_data *miiData=
                (struct mii_ioctl_data *)&(ifr->ifr_data);
            {
                miiData->val_out=Phy_GetRegW(pd,miiData->reg_num);
            }
        };break;
    case SIOCSMIIREG:
    case SIOCDEVPRIVATE+2:
        SMSC_TRACE(DBG_IOCTL,"SIOCSMIIREG");
        if(pd->LanInitialized) {
            struct mii_ioctl_data *miiData=
                (struct mii_ioctl_data *)&(ifr->ifr_data);
            {
                Phy_SetRegW(
                    pd,miiData->reg_num,((u16)(miiData->val_in)));
            }
        };break;
    case SIOCETHTOOL:
        result=Smsc9420_ethtool_ioctl(pd,userAddr);
        break;
    default:
        SMSC_WARNING("unknown cmd = 0x%08X",cmd);
        result=-EOPNOTSUPP;
        break;
    }
    
DONE:

    return result;
}





static void Lan_HandleGPTInt(pPD pd)
{
    u32 dwReadItBack=0;

        Lan_SetRegDW(INT_STAT,INT_STAT_GPT_INT_);
        dwReadItBack=Lan_GetRegDW(INT_STAT);
        return;
}
//NB: never attempt to reset MAC while PHY is in sleep
//doing so will keep mac in rst.
//following read or write  access to MAC register
//will almost certainly result in PCI read|write/retry infinite loop
//this will lock up the CPU for sure.
static void rst_mac_dmac(pPD pd)
{
    Lan_SetRegDW(BUS_MODE,BUS_MODE_SWR_);
    Lan_GetRegDW(BUS_MODE);
    udelay(MAX_MAC_DMAC_SRST_TIMEOUT_us);
    if(Lan_GetRegDW(BUS_MODE)&BUS_MODE_SWR_)
    {
        SMSC_WARNING("Software reset, BUS_MODE_SWR_ not cleared after %d us!",
                MAX_MAC_DMAC_SRST_TIMEOUT_us);
    }
    return;
}

static void config_gpio(pPD pd)
{
    if(GET_PD_DEBUG_MODE(pd) &
            DBG_ETH_GPIO)
    {
        //GPIO pins are used for debugging
        pd->GpoSetting=(GPIO_CFG_GPDIR2_ | GPIO_CFG_GPDIR1_ |
                GPIO_CFG_GPDIR0_ | GPIO_CFG_GPBUF2_ | GPIO_CFG_GPBUF1_ | 
                GPIO_CFG_GPBUF0_ );

    } else {
        //GPIO pins are driving LEDs
        pd->GpoSetting=(GPIO_CFG_LED_3_ | 
                GPIO_CFG_LED_2_ | GPIO_CFG_LED_1_ );
    }


    Lan_SetRegDW(GPIO_CFG,pd->GpoSetting);
    return;
}

static ulong Lan_Initialize(pPD pd)
{
    ulong result=0;
    SMSC_TRACE(DBG_INIT,"-->Lan_Initialize");
    //rst phy completely.
    //if the phy is in a power down mode
    //we cannot do a mac_rst;
    pmt_rst_phy(pd);
    //disable all other interrupts
    Lan_SetRegDW(INT_CTL,0UL);
    //Make sure interrupts are disabled
    //disable all DMAC interrupts
    Lan_SetRegDW(DMAC_INTR_ENA,0UL);
    //ack all the interrupts
    Lan_SetRegDW(INT_STAT,0xFFFFFFFFUL);
    //Do Reset
    rst_mac_dmac(pd);
    //make sure MAC_CR is sane
    Lan_SetRegDW(MAC_CR,0x0UL);

    //we need to rewrite mac address here
    Lan_SetRegDW(ADDRH,pd->MacHigh16);
    Lan_SetRegDW(ADDRL,pd->MacLow32);
    config_gpio(pd);


    //clear BUS_MODE BAR (Bus Arbitration)
    //we need to ensure that BAR bit in BUS_MODE is 0 before programing BUS_CFG 
    //with RXTXWeight, since we are writing it anyway, lets setup PBL (programmable
    //burst length
#ifdef __LITTLE_ENDIAN
    Lan_SetRegDW(BUS_MODE,(u32)(BUS_MODE_DMA_BURST_LENGTH_16));
#else
#ifdef __BIG_ENDIAN
    Lan_SetRegDW(BUS_MODE,(u32)(BUS_MODE_DMA_BURST_LENGTH_16 |
            BUS_MODE_DBO_ )
            );
#else
#error "Endiannes of the platforms is not defined"
#endif //BE
#endif //LE

    smsc9420_pci_push(pd);
    //set bus master bridge arbitration priority for Rx and TX DMA
    Lan_SetRegDW(BUS_CFG, CSR_RXTXWEIGHT_4_1);
    //we set SF and  OSF bits
    Lan_SetRegDW(DMAC_CONTROL,(u32)(DMAC_CONTROL_MBO_|
                DMAC_CONTROL_OSF_ ));        //SF_ store and forward should be used 

    pd->LanInitialized=1;
    smsc9420_pci_push(pd);


    result=1;
    SMSC_TRACE(DBG_INIT,"<--Lan_Initialize");
    return result;
}

static void Lan_SignalSoftwareInterrupt(pPD pd)
{
    pd->SoftwareInterruptSignal=0;
    Lan_EnableInterrupt(pd,INT_CTL_SW_INT_EN_);
}

static void Lan_HandleSWInt(pPD pd)
{
        Lan_DisableInterrupt(pd,INT_CTL_SW_INT_EN_);
        Lan_SetRegDW(INT_STAT,INT_STAT_SW_INT_);
        pd->SoftwareInterruptSignal=1;
        return;
}
static void Lan_HandleMBErrInt(pPD pd)
{
    //disable 
    Lan_DisableInterrupt(pd, INT_CTL_MBERR_INT_EN_); 
    //acknowledge
    Lan_SetRegDW(INT_STAT,INT_STAT_MBERR_INT_);
    //TODO
    //call software reset
    printk("MBERR Int\n");
    return;
}

static void Lan_HandleSBErrInt(pPD pd)
{
    //disable 
    Lan_DisableInterrupt(pd, INT_CTL_SBERR_INT_EN_); 
    //acknowledge
    Lan_SetRegDW(INT_STAT,INT_STAT_SBERR_INT_);
    //TODO
    //call software reset
    printk("SBERR Int\n");
    return;

}
static void Lan_EnableInterrupt(pPD pd,u32 dwIntEnMask)
{
    spinlock_int_t dwIntFlags=0;
    spin_lock_irqsave(&(pd->INT_CTL_lock),dwIntFlags);
    {
        Lan_SetRegDW(INT_CTL,Lan_GetRegDW(INT_CTL)|dwIntEnMask);
    }
    spin_unlock_irqrestore(&(pd->INT_CTL_lock),dwIntFlags);
}

static void Lan_DisableInterrupt(pPD pd,u32 dwIntEnMask)
{
   spinlock_int_t  dwIntFlags=0;
    spin_lock_irqsave(&(pd->INT_CTL_lock),dwIntFlags);
    {
        Lan_SetRegDW(INT_CTL,Lan_GetRegDW(INT_CTL)&(~dwIntEnMask));
    }
    spin_unlock_irqrestore(&(pd->INT_CTL_lock),dwIntFlags);
}

static void Lan_ClearDeassertionTimer(pPD pd)
{
    spinlock_int_t dwIntFlags=0;
    spin_lock_irqsave(&(pd->INT_CFG_lock),dwIntFlags);
    {
        Lan_SetRegDW(INT_CFG,Lan_GetRegDW(INT_CFG)|INT_CFG_INT_DEAS_CLR_);
    }
    spin_unlock_irqrestore(&(pd->INT_CFG_lock),dwIntFlags);
}


static void Lan_Disable_RXDMACInt_nolock(pPD pd)
{
    u32 dma_intr_ena = Lan_GetRegDW(DMAC_INTR_ENA);
    //CLEAR RXBU_, RX_
    dma_intr_ena &= ~(DMAC_INTR_ENA_RXBU_|
            DMAC_INTR_ENA_RX_);

    Lan_SetRegDW(DMAC_INTR_ENA,dma_intr_ena);
    return;
}

static void Lan_Disable_TXDMACInt_nolock(pPD pd)
{
    u32 dma_intr_ena = Lan_GetRegDW(DMAC_INTR_ENA);
    //CLEAR TX_, TXBU_
    dma_intr_ena &= ~(DMAC_INTR_ENA_TX_ |
            DMAC_INTR_ENA_TXBU_);
    Lan_SetRegDW(DMAC_INTR_ENA, dma_intr_ena);
    smsc9420_pci_push(pd);
    return;
}

static void Lan_Enable_TXDMACInt_nolock(pPD pd)
{
    u32 dma_intr_ena = Lan_GetRegDW(DMAC_INTR_ENA);
    //do not enable TXBU_, just TX_
    dma_intr_ena |= (DMAC_INTR_ENA_TX_ |
                     DMAC_INTR_ENA_NIS_|
#ifdef LAN94200_USE_AIS
                    DMAC_INTR_ENA_AIS_
#else
                     0
#endif
                     );
    Lan_SetRegDW(DMAC_INTR_ENA, dma_intr_ena);
    smsc9420_pci_push(pd);
    return;
}
static void Lan_Enable_RXDMACInt_nolock(pPD pd)
{
    u32 dma_intr_ena = Lan_GetRegDW(DMAC_INTR_ENA);
    dma_intr_ena |= (DMAC_INTR_ENA_RX_ |
                     DMAC_INTR_ENA_NIS_| 
#ifdef LAN94200_USE_AIS
                    DMAC_INTR_ENA_AIS_
#else
                    0
#endif
            );

    Lan_SetRegDW(DMAC_INTR_ENA,dma_intr_ena);
    smsc9420_pci_push(pd);
    return;

}

static void Lan_EnableIRQ(pPD pd)
{
    spinlock_int_t dwIntFlags=0;
    spin_lock_irqsave(&(pd->INT_CFG_lock),dwIntFlags);
            Lan_SetRegDW(INT_CFG,Lan_GetRegDW(INT_CFG)|INT_CFG_IRQ_EN_);
    spin_unlock_irqrestore(&(pd->INT_CFG_lock),dwIntFlags);
}

static void Lan_DisableIRQ(pPD pd)
{
    spinlock_int_t dwIntFlags=0;
    spin_lock_irqsave(&(pd->INT_CFG_lock),dwIntFlags);
            Lan_SetRegDW(INT_CFG,(u32)(Lan_GetRegDW(INT_CFG)&(~INT_CFG_IRQ_EN_)));
    spin_unlock_irqrestore(&(pd->INT_CFG_lock),dwIntFlags);
}
static void Lan_SetIntDeas(pPD pd, u32 dwIntDeas)
{
    spinlock_int_t dwIntFlags=0;
    spin_lock_irqsave(&(pd->INT_CFG_lock),dwIntFlags);
    {
        u32 dwIntCfg=Lan_GetRegDW(INT_CFG);
        dwIntCfg&=(~(INT_CFG_INT_DEAS_MASK));
        dwIntCfg|=(dwIntDeas&INT_CFG_INT_DEAS_MASK);
        Lan_SetRegDW(INT_CFG,dwIntCfg);
    }
    spin_unlock_irqrestore(&(pd->INT_CFG_lock),dwIntFlags);
}


//NB: Milos does not support external PHY
//this makes our life easy. All Phy registers are accessible
static u16 Phy_GetRegW(
    pPD pd,
    u32 dwRegIndex)
{
    u32 dwAddr=0;
    u16 result=(~0);
    spinlock_int_t flags;
    ulong max_waitcycle_count = 
        (PHY_ACCESS_MAX_WAIT_us/PHY_ACCESS_DELAY_TICK_us);

    spin_lock_irqsave(&pd->phy_lock, flags);

    // confirm MII not busy
    if((Lan_GetRegDW(MII_ACCESS)&MII_ACCESS_MII_BUSY_)!=0)
    {
        SMSC_WARNING("MII is busy???");
        goto DONE;
    }

    //set the address, index & direction (read from PHY)
    dwAddr = ((pd->dwPhyAddress&0x1FUL)<<11)|
            ((dwRegIndex&0x1F)<<6)|MII_ACCESS_MII_READ_;
    Lan_SetRegDW(MII_ACCESS,dwAddr);

    // wait for read to complete w/ timeout
    do
    {
        if(!(max_waitcycle_count--))
        {
            goto DONE;
        }
        udelay(PHY_ACCESS_DELAY_TICK_us);

    }while(Lan_GetRegDW(MII_ACCESS)&MII_ACCESS_MII_BUSY_);
    if(!max_waitcycle_count)
    {
        SMSC_WARNING("MII busy timeout!");
    }

    result = (u16)Lan_GetRegDW(MII_DATA);


DONE:
    spin_unlock_irqrestore(&pd->phy_lock, flags);
    return result;
}

//NB: Milos does not support external PHY
//this makes our life easy. All Phy registers are accessible
static void Phy_SetRegW(
    pPD pd,
    u32 dwRegIndex,u16 wVal)
{

    u32 dwAddr=0;
    spinlock_int_t flags;
    ulong max_waitcycle_count=
        (PHY_ACCESS_MAX_WAIT_us/PHY_ACCESS_DELAY_TICK_us);


    spin_lock_irqsave(&pd->phy_lock, flags);

    if(dwRegIndex==0) {
        if((wVal&(PHY_BCR_ANEG_EN_|PHY_BCR_RST_ANEG_))==
            (PHY_BCR_ANEG_EN_|PHY_BCR_RST_ANEG_))
        {
            pd->wLastADVatRestart=pd->wLastADV;
        }
    }
    if(dwRegIndex==4) {
        pd->wLastADV=wVal;
    }
    
    // confirm MII not busy
    if((Lan_GetRegDW(MII_ACCESS)&MII_ACCESS_MII_BUSY_)!=0UL)
    {
        SMSC_WARNING("MII is busy???");
        goto DONE;
    }

    //put the data to write in the MAC
    Lan_SetRegDW(MII_DATA,(u32)wVal);

    //set the address, index & direction (write to PHY)
    dwAddr = ((pd->dwPhyAddress&0x1FUL)<<11) |
            ((dwRegIndex&0x1FUL)<<6)|MII_ACCESS_MII_WRITE_;
    Lan_SetRegDW(MII_ACCESS,dwAddr);

    //wait for write to complete w/ timeout
    do
    {
        if(!(max_waitcycle_count--))
        {
            goto DONE;
        }
        udelay(PHY_ACCESS_DELAY_TICK_us);

    }while(Lan_GetRegDW(MII_ACCESS)&MII_ACCESS_MII_BUSY_);
DONE:
    spin_unlock_irqrestore(&pd->phy_lock, flags);
    return;
}
static void pmt_rst_phy(pPD pd)
{
    ulong phy_to = 1000;  //we sleep for 10us on each itteration
        //reset PHY completely
        Lan_SetRegDW(PMT_CTRL, 
                (PMT_CTRL_PHY_RST_ |
                Lan_GetRegDW(PMT_CTRL)) );
        while(--phy_to)
        {
            udelay(10);
            if(!(PMT_CTRL_PHY_RST_ &
                        Lan_GetRegDW(PMT_CTRL)))
            {
                break;
            }
        }
        if(!phy_to)
        {
            SMSC_WARNING("failed to reset PHY: %d\n", __LINE__);
        }
        return;
}

static ulong Phy_Initialize(pPD pd, u32 dwLinkRequest)
{
    ulong result=0;
    u32 dwPhyId=0;
    u16 wTemp=0;
    ulong dwLoopCount=0;
    

    pd->dwPhyAddress=1UL;

    pmt_rst_phy(pd);

    //no need to acquire phy lock because this is 
    //  part of the single threaded initialization path
    dwPhyId=Phy_GetRegW(pd,PHY_ID1);
    pd->Phy_OUI=((dwPhyId&0xFFFF)<< 6);

    dwPhyId=Phy_GetRegW(pd,PHY_ID2);
   
    pd->Phy_OUI |= ((dwPhyId & 0xFC00) >> 10);

    pd->Phy_Rev=((u8)(dwPhyId&(0x0FUL)));

    pd->Phy_Model=((u8)((dwPhyId>>4)&(0x3FUL)));
    
    SMSC_TRACE(DBG_INIT,"Phy_OUI==0x%08X,Phy_Model==0x%02X,Phy_Rev==0x%02X",
        pd->Phy_OUI,
        pd->Phy_Model,
        pd->Phy_Rev);

    pd->dwLinkSpeed=LINK_OFF;
    pd->dwLinkSettings=LINK_OFF;
    //reset the PHY
    Phy_SetRegW(pd,PHY_BCR,PHY_BCR_RESET_);

    smsc9420_pci_push(pd);
    dwLoopCount=100000;

    do {
        udelay(10);
        wTemp=Phy_GetRegW(pd,PHY_BCR);
        dwLoopCount--;
    } while((dwLoopCount>0)&&(wTemp&PHY_BCR_RESET_));

    
    if(wTemp&PHY_BCR_RESET_) {
        SMSC_WARNING("PHY reset failed to complete.");
        goto DONE;
    }

    // wait 100 msec after reset then set link
    dwLoopCount=100;
    do {
        udelay(1000);
        dwLoopCount--;
    } while(dwLoopCount>0);
    
    Phy_SetLink(pd,dwLinkRequest);

    init_timer(&(pd->LinkPollingTimer));
    pd->StopLinkPolling=0;
    pd->LinkPollingTimer.function=Phy_CheckLink;
    pd->LinkPollingTimer.data=(unsigned long)pd;
    pd->LinkPollingTimer.expires=jiffies+HZ;
    add_timer(&(pd->LinkPollingTimer));

    result=1;

DONE:
    SMSC_TRACE(DBG_INIT,"<--Phy_Initialize");
    return result;
}


static void Phy_SetLink(pPD pd,
                 u32 dwLinkRequest) 
{
    if(dwLinkRequest&LINK_AUTO_NEGOTIATE) {
        u16 wTemp;
        wTemp=Phy_GetRegW(pd,
            PHY_ANEG_ADV);
        wTemp&=~PHY_ANEG_ADV_PAUSE_;
        if(dwLinkRequest&LINK_ASYMMETRIC_PAUSE) {
            wTemp|=PHY_ANEG_ADV_ASYMP_;
        } 
        if(dwLinkRequest&LINK_SYMMETRIC_PAUSE) {
            wTemp|=PHY_ANEG_ADV_SYMP_;
        }
        wTemp&=~PHY_ANEG_ADV_SPEED_;
        if(dwLinkRequest&LINK_SPEED_10HD) {
            wTemp|=PHY_ANEG_ADV_10H_;
        }
        if(dwLinkRequest&LINK_SPEED_10FD) {
            wTemp|=PHY_ANEG_ADV_10F_;

        }
        if(dwLinkRequest&LINK_SPEED_100HD) {
            wTemp|=PHY_ANEG_ADV_100H_;
        }
        if(dwLinkRequest&LINK_SPEED_100FD) {
            wTemp|=PHY_ANEG_ADV_100F_;
        }
        Phy_SetRegW(pd,PHY_ANEG_ADV,wTemp);

        // begin to establish link
        Phy_SetRegW(pd,
            PHY_BCR,
            PHY_BCR_ANEG_EN_|
            PHY_BCR_RST_ANEG_);
    } else {
        u16 wTemp=0;
        if(dwLinkRequest&(LINK_SPEED_100FD)) {
            dwLinkRequest=LINK_SPEED_100FD;
        } else if(dwLinkRequest&(LINK_SPEED_100HD)) {
            dwLinkRequest=LINK_SPEED_100HD;
        } else if(dwLinkRequest&(LINK_SPEED_10FD)) {
            dwLinkRequest=LINK_SPEED_10FD;
        } else if(dwLinkRequest&(LINK_SPEED_10HD)) {
            dwLinkRequest=LINK_SPEED_10HD;
        }
        if(dwLinkRequest&(LINK_SPEED_10FD|LINK_SPEED_100FD)) {
            wTemp|=PHY_BCR_DUPLEX_MODE_;
        }
        if(dwLinkRequest&(LINK_SPEED_100HD|LINK_SPEED_100FD)) {
            wTemp|=PHY_BCR_SPEED_SELECT_;
        }
        Phy_SetRegW(pd,PHY_BCR,wTemp);
    }
    return;
}

static void Phy_CheckLink(unsigned long ptr)
{

    pPD pd=(pPD)ptr;
    if(pd==NULL) {
        SMSC_WARNING("ptr==0");
        return;
    }

    if(!pd->LanInitialized)
    {
        return;
    }
    //must call this twice
    Phy_UpdateLinkMode(pd);
    Phy_UpdateLinkMode(pd);

    if(!(pd->StopLinkPolling)) {
        pd->LinkPollingTimer.expires=jiffies+HZ;
        add_timer(&(pd->LinkPollingTimer));
    }
    return;
}

static void Phy_UpdateLinkMode(pPD pd)
{
    u32 dwOldLinkSpeed=pd->dwLinkSpeed;

    Phy_GetLinkMode(pd);


    if(dwOldLinkSpeed!=(pd->dwLinkSpeed)) {
        if(pd->dwLinkSpeed!=LINK_OFF) {
            u32 dwRegVal=0;
            switch(pd->dwLinkSpeed) {
            case LINK_SPEED_10HD:
                SMSC_TRACE(DBG_LINK,"Link is now UP at 10Mbps HD");
                break;
            case LINK_SPEED_10FD:
                SMSC_TRACE(DBG_LINK,"Link is now UP at 10Mbps FD");
                break;
            case LINK_SPEED_100HD:
                SMSC_TRACE(DBG_LINK,"Link is now UP at 100Mbps HD");
                break;
            case LINK_SPEED_100FD:
                SMSC_TRACE(DBG_LINK,"Link is now UP at 100Mbps FD");
                break;

            default:
                SMSC_WARNING("Link is now UP at Unknown Link Speed, dwLinkSpeed=0x%08X",
                    pd->dwLinkSpeed);
                break;
            }
        
            dwRegVal=Lan_GetRegDW(MAC_CR);
            dwRegVal&=~(MAC_CR_FDPX_|MAC_CR_DIS_RXOWN_);

            switch(pd->dwLinkSpeed) {
            case LINK_SPEED_10HD:
            case LINK_SPEED_100HD:
                dwRegVal|=MAC_CR_DIS_RXOWN_;
                break;
            case LINK_SPEED_10FD:
            case LINK_SPEED_100FD:
                dwRegVal|=MAC_CR_FDPX_;
                break;
            default:break;//make lint happy
            }

            Lan_SetRegDW(MAC_CR,dwRegVal);

            if(pd->dwLinkSettings&LINK_AUTO_NEGOTIATE) {
                u16 linkPartner=0;
                u16 localLink=0;
                localLink=Phy_GetRegW(pd,PHY_ANEG_ADV);
                linkPartner=Phy_GetRegW(pd,PHY_ANEG_LPA);
                switch(pd->dwLinkSpeed) {
                case LINK_SPEED_10FD:
                case LINK_SPEED_100FD:
                    if(((localLink&linkPartner)&((u16)0x0400U)) != ((u16)0U)) {
                        //Enable PAUSE receive and transmit
                        Lan_SetRegDW(FLOW,0xFFFF0002UL);
                    
                    } else if(((localLink&((u16)0x0C00U))==((u16)0x0C00U)) &&
                            ((linkPartner&((u16)0x0C00U))==((u16)0x0800U)))
                    {
                        //Enable PAUSE receive, disable PAUSE transmit
                        Lan_SetRegDW(FLOW,0xFFFF0002UL);
                    
                    } else {
                        //Disable PAUSE receive and transmit
                        Lan_SetRegDW(FLOW,0UL);
                    
                    };break;
                case LINK_SPEED_10HD:
                case LINK_SPEED_100HD:
                    Lan_SetRegDW(FLOW,0UL);
                    
                    break;
                default:break;//make lint happy
                }
                SMSC_TRACE(DBG_LINK,"LAN9420: %s,%s,%s,%s,%s,%s",
                    (localLink&PHY_ANEG_ADV_ASYMP_)?"ASYMP":"     ",
                    (localLink&PHY_ANEG_ADV_SYMP_)?"SYMP ":"     ",
                    (localLink&PHY_ANEG_ADV_100F_)?"100FD":"     ",
                    (localLink&PHY_ANEG_ADV_100H_)?"100HD":"     ",
                    (localLink&PHY_ANEG_ADV_10F_)?"10FD ":"     ",
                    (localLink&PHY_ANEG_ADV_10H_)?"10HD ":"     ");
    
                SMSC_TRACE(DBG_LINK,"Partner: %s,%s,%s,%s,%s,%s",

                    (linkPartner&PHY_ANEG_LPA_ASYMP_)?"ASYMP":"     ",
                    (linkPartner&PHY_ANEG_LPA_SYMP_)?"SYMP ":"     ",
                    (linkPartner&PHY_ANEG_LPA_100FDX_)?"100FD":"     ",
                    (linkPartner&PHY_ANEG_LPA_100HDX_)?"100HD":"     ",
                    (linkPartner&PHY_ANEG_LPA_10FDX_)?"10FD ":"     ",
                    (linkPartner&PHY_ANEG_LPA_10HDX_)?"10HD ":"     ");
            } else {
                switch(pd->dwLinkSpeed) {
                case LINK_SPEED_10HD:
                case LINK_SPEED_100HD:
                    Lan_SetRegDW(FLOW,0x0UL);
                    
                    break;
                default:
                    Lan_SetRegDW(FLOW,0x0UL);
                    
                    break;
                }
            }
            smsc9420_pci_push(pd);
            netif_carrier_on(GET_PD_DEV(pd));
            
        } else {
            SMSC_TRACE(DBG_LINK,"Link is now DOWN");
//            netif_tx_disable(GET_PD_DEV(pd));
            netif_carrier_off(GET_PD_DEV(pd));
            Lan_SetRegDW(FLOW,0UL);
        }
    }
    return;
}

static void Phy_GetLinkMode(
    pPD pd)
{
    ulong result=LINK_OFF;
    u16 wRegVal=0;
    u16 wRegBSR=Phy_GetRegW(
        pd,
        PHY_BSR);
    pd->dwLinkSettings=LINK_OFF;
    if(wRegBSR&PHY_BSR_LINK_STATUS_) {
        wRegVal=Phy_GetRegW(
            pd,
            PHY_BCR);
        if(wRegVal&PHY_BCR_ANEG_EN_) {
            ulong  linkSettings=LINK_AUTO_NEGOTIATE;
            u16 wRegADV=
                pd->wLastADVatRestart;

            u16 wRegLPA=Phy_GetRegW(
                pd,
                PHY_ANEG_LPA);
            if(wRegADV&PHY_ANEG_ADV_ASYMP_) {
                linkSettings|=LINK_ASYMMETRIC_PAUSE;
            }
            if(wRegADV&PHY_ANEG_ADV_SYMP_) {
                linkSettings|=LINK_SYMMETRIC_PAUSE;
            }
            if(wRegADV&PHY_ANEG_LPA_100FDX_) {
                linkSettings|=LINK_SPEED_100FD;
            }
            if(wRegADV&PHY_ANEG_LPA_100HDX_) {
                linkSettings|=LINK_SPEED_100HD;
            }
            if(wRegADV&PHY_ANEG_LPA_10FDX_) {
                linkSettings|=LINK_SPEED_10FD;

            }
            if(wRegADV&PHY_ANEG_LPA_10HDX_) {
                linkSettings|=LINK_SPEED_10HD;
            }
            pd->dwLinkSettings=linkSettings;
            wRegLPA&=wRegADV;
            if(wRegLPA&PHY_ANEG_LPA_100FDX_) {
                result=LINK_SPEED_100FD;
            } else if(wRegLPA&PHY_ANEG_LPA_100HDX_) {
                result=LINK_SPEED_100HD;
            } else if(wRegLPA&PHY_ANEG_LPA_10FDX_) {
                result=LINK_SPEED_10FD;
            } else if(wRegLPA&PHY_ANEG_LPA_10HDX_) {
                result=LINK_SPEED_10HD;
            }
        } else {
            if(wRegVal&PHY_BCR_SPEED_SELECT_) {
                if(wRegVal&PHY_BCR_DUPLEX_MODE_) {
                    pd->dwLinkSettings=result=LINK_SPEED_100FD;
                } else {
                    pd->dwLinkSettings=result=LINK_SPEED_100HD;
                }
            } else {
                if(wRegVal&PHY_BCR_DUPLEX_MODE_) {
                    pd->dwLinkSettings=result=LINK_SPEED_10FD;
                } else {
                    pd->dwLinkSettings=result=LINK_SPEED_10HD;
                }
            }
        }
    }
    pd->dwLinkSpeed=result;
    return;
}

static ulong Tx_Initialize(pPD pd)
{
    ulong result=0;
    ulong index=0;

    //this need not be aligned
    pd->TxSkBuffers=
        kmalloc((sizeof(INFO_TX_SkBUFFERS)*pd->tx_ring_size),
                GFP_KERNEL);



    if(pd->TxSkBuffers==NULL) {
        SMSC_WARNING("Failed to allocated TxSkBuffers");
        goto DONE;
    }

    
    //Initialize the TX Ring
    for(index=0;index<pd->tx_ring_size;index++) 
    {
        pd->TxSkBuffers[index].TxSkb=NULL;
        pd->TxSkBuffers[index].mapping_TxSkBuffers=0;
        pd->TxSkBuffers[index].len=0;
        pd->TxDescRing[index].dwStatus=(u32)(0);
        pd->TxDescRing[index].dwLength=(u32)(0);
        pd->TxDescRing[index].dwBuffer1=(u32)(0);
        pd->TxDescRing[index].dwBuffer2=(u32)(0);
    }
    pd->TxDescRing[pd->tx_ring_size-1].dwLength=
        (u32)(TDES1_TER_);

    wmb();
    TX_SET_HEAD(pd, 0);
    TX_SET_TAIL(pd, 0);
    TX_CLR_FULL(pd);

    

    Lan_SetRegDW(TX_BASE_ADDR,PD_GET_TXDMA_ADDR(pd));



    result=1;
    smsc9420_pci_push(pd);
DONE:
    if(result!=1) {
        Tx_CleanUp(pd);
    }
    return result;
}


static void Tx_hw_start(pPD pd)
{
    Lan_SetRegDW(MAC_CR,
       ( Lan_GetRegDW(MAC_CR))|
       (MAC_CR_TXEN_));

    Lan_SetRegDW(DMAC_CONTROL,
       ( Lan_GetRegDW(DMAC_CONTROL))|
       DMAC_CONTROL_ST_);

    smsc9420_pci_push(pd);

    //we are executing with IRQ disabled
    Lan_Enable_TXDMACInt_nolock(pd);

    return;
}
static void Tx_hw_stop(pPD pd)
{
    //stop the transmitter. Wait for idle state for a maximum of 10 ms.
    //A max size ethernet frame at 10 Mb/s would take about 1.5ms to Xfer,
    //so with 10ms it should be enough time to complete any transmit in
    //progress (DMA is out to the Tx buffer and Tx on the wire)
    ulong timeOut=1000; //1000 * 10us -> 10ms
    //disable TX DMAC
    Lan_SetRegDW(DMAC_CONTROL,
            (u32)(Lan_GetRegDW(DMAC_CONTROL)&(~DMAC_CONTROL_ST_)));
    //Wait for transmit process to stop
    while((timeOut)&&(Lan_GetRegDW(DMAC_STATUS)&DMAC_STS_TS_))
    {
       udelay(10);
       timeOut--;
    }
    if(timeOut==0) 
    {
       SMSC_WARNING("TX DMAC failed to stop! Timeout.");
    }
    //ACK Tx DMAC stop bit
    Lan_SetRegDW(DMAC_STATUS, DMAC_STS_TXPS_);
        
    Lan_Disable_TXDMACInt_nolock(pd);
    //stop MAC TX 
    Lan_SetRegDW(MAC_CR,
            (Lan_GetRegDW(MAC_CR)&(~MAC_CR_TXEN_)));
    smsc9420_pci_push(pd);
    return;
}

static void Tx_buffers_destroy(pPD pd)
{
    ulong index;
    if(!pd->TxSkBuffers) 
    {
        return;
    }

    for(index=0;index<pd->tx_ring_size;index++) 
    {
                
        //destroy PCI-DMA mappings
      if (pd->TxSkBuffers[index].mapping_TxSkBuffers) 
      {
        pci_unmap_single(GET_PD_PDEV(pd),
                pd->TxSkBuffers[index].mapping_TxSkBuffers,
                pd->TxSkBuffers[index].len,
                PCI_DMA_TODEVICE);
      }
      //if there is an skb free it
      if(pd->TxSkBuffers[index].TxSkb)
      {
         dev_kfree_skb_any(pd->TxSkBuffers[index].TxSkb);
      }
      //mark descriptors
      if(pd->TxDescRing)
      {
        pd->TxDescRing[index].dwStatus = 0;
        pd->TxDescRing[index].dwLength= 0;
        pd->TxDescRing[index].dwBuffer1= 0;
        pd->TxDescRing[index].dwBuffer2= 0;
      }
    }
    kfree(pd->TxSkBuffers);
    pd->TxHeadIndex = 0;
    pd->TxTailIndex = 0;
    pd->TxFull = 0;
    pd->TxSkBuffers=NULL;
    return;
}
static void Tx_CleanUp(pPD pd)
{
   Tx_hw_stop(pd);
   Tx_buffers_destroy(pd);
   return;
}


static ulong Rx_Initialize(pPD pd)
{
    ulong result=0;
    ulong index=0;
    pd->RxSkBuffers=
        kmalloc((sizeof(INFO_RX_SkBUFFERS)*pd->rx_ring_size),
                GFP_KERNEL);
    if(pd->RxSkBuffers==NULL) {
        SMSC_WARNING("Failed to allocated RxSkBuffers");
        goto DONE;
    }

    //Initialize the RX Ring
    for(index=0;index<pd->rx_ring_size;index++)
    {
        pd->RxDescRing[index].dwStatus = (u32)(0);
        pd->RxDescRing[index].dwLength = (u32)(PKT_BUF_SZ);
        pd->RxDescRing[index].dwBuffer2 = (u32)(0);
        pd->RxSkBuffers[index].RxSkb=NULL;
        pd->RxSkBuffers[index].mapping_RxSkBuffers=0;
    }
    //Wraparound
    pd->RxDescRing[index-1].dwLength = (u32)(PKT_BUF_SZ| RDES1_RER_);

    //Now allocate the Rx SkBuffers
    for(index=0; index< pd->rx_ring_size;index++)
    {
        dma_addr_t mapping;
        struct sk_buff * skb;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
        skb    = dev_alloc_skb(PKT_BUF_SZ);
#else
        skb    = netdev_alloc_skb(GET_PD_DEV(pd),PKT_BUF_SZ);
#endif
        pd->RxSkBuffers[index].RxSkb=skb;
        if(skb==NULL) {
            SMSC_WARNING("Failed to Allocated Rx SKB, index=%lu",index);
            goto DONE;
        }
        #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22))
        mapping=pci_map_single(GET_PD_PDEV(pd), 
                (ulong *)skb->tail, PKT_BUF_SZ, PCI_DMA_FROMDEVICE);
        #else
        mapping=pci_map_single(GET_PD_PDEV(pd), 
                (ulong *)skb_tail_pointer(skb), PKT_BUF_SZ, PCI_DMA_FROMDEVICE);
        #endif
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27))
        if(pci_dma_mapping_error(mapping))
#else
        if(pci_dma_mapping_error(GET_PD_PDEV(pd), mapping))
#endif
        {
            dev_kfree_skb_any(skb);
            pd->RxSkBuffers[index].RxSkb = NULL;
            SMSC_WARNING("pci_map_single failed!");
            goto DONE;
        }
        pd->RxSkBuffers[index].mapping_RxSkBuffers=mapping;
        skb->dev = GET_PD_DEV(pd);
        pd->RxDescRing[index].dwBuffer1=(u32)((u32)mapping+
                (u32)NET_IP_ALIGN);
        pd->RxDescRing[index].dwStatus=(u32)(RDES0_OWN_);
    }


    pd->RxHeadIndex=0;
    pd->RxTailIndex=0;
    pd->RxRingFull=0;

    Lan_SetRegDW(VLAN1, ETH_P_8021Q);
    SMSC_TRACE(DBG_INIT,"VLAN1 = 0x%08lx\n", (ulong)Lan_GetRegDW(VLAN1));

    if (IS_PD_RxCSUM_SET(pd))  {
        
            //Set RX COE

            #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15))
                Lan_SetRegDW(COE_CR, Lan_GetRegDW(COE_CR) | RX_COE_EN | RX_COE_MODE);
            #else
                Lan_SetRegDW(COE_CR, Lan_GetRegDW(COE_CR) | RX_COE_EN );
            #endif

            SMSC_TRACE(DBG_INIT,"COE_CR = 0x%08lx\n", (ulong)Lan_GetRegDW(COE_CR));

        
    }

    Lan_SetRegDW(RX_BASE_ADDR,pd->rx_dma_addr);


    smsc9420_pci_push(pd);
    result=1;
DONE:
    if(result!=1) {
        Rx_CleanUp(pd);
    }
    return result;
}

static void Rx_hw_start(pPD pd)
{
    Lan_SetRegDW(DMAC_CONTROL,
        (u32)(Lan_GetRegDW(DMAC_CONTROL))|
        (u32)(DMAC_CONTROL_SR_));
    Lan_SetRegDW(MAC_CR,
        (Lan_GetRegDW(MAC_CR))|
        MAC_CR_RXEN_);
    smsc9420_pci_push(pd);
    //we are executing with IRQ disabled
    Lan_Enable_RXDMACInt_nolock(pd);

    return;
}
static void Rx_hw_stop(pPD pd)
{

      int timeOut=1000; //1000 * 10us = 10ms
      Lan_Disable_RXDMACInt_nolock(pd);
    //stop RX MAC prior to stoping DMA
      Lan_SetRegDW(MAC_CR,
           (u32)( Lan_GetRegDW(MAC_CR)&(~MAC_CR_RXEN_)));
      //push PCI transaction
      smsc9420_pci_push(pd);
      //stop RX DMAC
      Lan_SetRegDW(DMAC_CONTROL,
            (u32)(Lan_GetRegDW(DMAC_CONTROL)&(~DMAC_CONTROL_SR_)));
      //push PCI transaction
      smsc9420_pci_push(pd);

      //Wait for Receive Process to stop

      //wait for idle state for a maximum of 10ms
      //A max size ethernet frame at 10mb/s would take about 1,5 ms
      //to Xfer. so with 10ms it should be enough time to complete any
      //receive in progress (Rx it into the buffer + DMA it into the host memory)
      while((timeOut)&&(Lan_GetRegDW(DMAC_STATUS)&DMAC_STS_RS_)) {
          udelay(10);
          timeOut--;
      }
      if(timeOut==0) {
           SMSC_WARNING("RX DMAC did not stop! timeout.");
       }
      //ACK the Rx DMAC stop bit
      Lan_SetRegDW(DMAC_STATUS, DMAC_STS_RXPS_);
      return;
}
    
static void Rx_buffers_destroy(pPD pd)
{
    ulong index;

    if(!pd->RxSkBuffers) 
    {
       return;
    }
    //go through the entire ring
    for(index=0;index<pd->rx_ring_size;index++) 
    {

        if(pd->RxSkBuffers[index].RxSkb) 
        {
            //destroy skb
            dev_kfree_skb_any(pd->RxSkBuffers[index].RxSkb);
        }
        if(pd->RxSkBuffers[index].mapping_RxSkBuffers) 
        {
            //destroy PCI-DMA mapping 
            pci_unmap_single(GET_PD_PDEV(pd), 
                    pd->RxSkBuffers[index].mapping_RxSkBuffers,
                    PKT_BUF_SZ, PCI_DMA_FROMDEVICE);
            pd->RxSkBuffers[index].mapping_RxSkBuffers = 0;
        }
        if(pd->RxDescRing)
        {
            pd->RxDescRing[index].dwStatus = 0;
            pd->RxDescRing[index].dwLength= 0;
            pd->RxDescRing[index].dwBuffer1= 0;
            pd->RxDescRing[index].dwBuffer2= 0;
        }
            
    }

    kfree(pd->RxSkBuffers);
    pd->RxHeadIndex = 0;
    pd->RxTailIndex = 0;
    pd->RxSkBuffers=NULL;
    return;
}

static void Rx_CleanUp(pPD pd)
{
    Rx_hw_stop(pd);
    Rx_buffers_destroy(pd);
    return;
}

static ulong Rx_IncrementHeadIndex(pPD pd)
{
    SMSC_ASSERT(pd!=NULL);

    if((pd->RxHeadIndex)==(pd->RxTailIndex))
    {
        if(pd->RxRingFull) {
            return 0;
        }
    } else {
        SMSC_ASSERT(pd->RxRingFull==0);
    }
    pd->RxHeadIndex++;
    if(pd->RxHeadIndex>=pd->rx_ring_size) {
        SMSC_ASSERT(pd->RxHeadIndex==pd->rx_ring_size);
        pd->RxHeadIndex=0;
    }
    if(pd->RxHeadIndex==pd->RxTailIndex) {
        pd->RxRingFull=1;
    }
    return 1;
}

static ulong Rx_IncrementTailIndex(pPD pd)
{
    SMSC_ASSERT(pd!=NULL);
    if((pd->RxHeadIndex)==(pd->RxTailIndex)) {
        if(!(pd->RxRingFull)) {
            return 0;
        }
    } else {
        SMSC_ASSERT(pd->RxRingFull==0);
    }
    pd->RxRingFull=0;
    pd->RxTailIndex++;
    if(pd->RxTailIndex>=pd->rx_ring_size) {
        SMSC_ASSERT(pd->RxTailIndex==pd->rx_ring_size);
        pd->RxTailIndex=0;
    }
    return 1;
}

static ulong Rx_GetFreeDescriptorCount(pPD pd)
{
    ulong result=0;

    SMSC_ASSERT(pd!=NULL);
    SMSC_ASSERT(pd->RxTailIndex<pd->rx_ring_size);
    SMSC_ASSERT(pd->RxHeadIndex<pd->rx_ring_size);
    if((pd->RxHeadIndex)==(pd->RxTailIndex)) {
        if(pd->RxRingFull) {
            result=0;
        } else {
            result= pd->rx_ring_size;
        }
    } else if((pd->RxHeadIndex)>(pd->RxTailIndex)) {
        result=pd->rx_ring_size-pd->RxHeadIndex+
                            pd->RxTailIndex;
    } else {
        result=(pd->RxTailIndex)-(pd->RxHeadIndex);
    }

    return result;
}


//returns hash bit number for given MAC address
static u32 Rx_Hash(u8 addr[6])
{
    int i;
    u32 crc=0xFFFFFFFFUL;

    u32 poly=0xEDB88320UL;
    u32 result=0;
    for(i=0;i<6;i++)
    {
        int bit;
        u32 data=((u32)addr[i]);
        for(bit=0;bit<8;bit++)
        {
            u32 p = (crc^((u32 )data))&1UL;
            crc >>= 1;
            if(p!=0) crc ^= poly;
            data >>=1;
        }
    }
    result=((crc&0x01UL)<<5)|
        ((crc&0x02UL)<<3)|
        ((crc&0x04UL)<<1)|
        ((crc&0x08UL)>>1)|
        ((crc&0x10UL)>>3)|
        ((crc&0x20UL)>>5);
    return result;
}

static void Rx_SetMulticastList(pPD pd)
{
    u32 rxFilter=0;
    u32 rxHashLo=0;
    u32 rxHashHi=0;
    SMSC_ASSERT(pd!=NULL);
    SMSC_ASSERT(IS_PD_DEV_SET(pd));
    rxFilter=pd->RxFoxFilter;
    rxHashLo=pd->RxFoxHashLo;
    rxHashHi=pd->RxFoxHashHi;
    
    if(rxFilter&RX_FILTER_PROMISCUOUS) {
        u32 dwMacCr=Lan_GetRegDW(MAC_CR);
        dwMacCr|=MAC_CR_PRMS_;
        dwMacCr&=(~MAC_CR_MCPAS_);
        dwMacCr&=(~MAC_CR_HPFILT_);
        Lan_SetRegDW(MAC_CR,dwMacCr);
        smsc9420_pci_push(pd);
        SMSC_TRACE(DBG_MCAST, "Promiscuous Mode Enabled");
        return;
    }
    if(rxFilter&RX_FILTER_ALL_MULTICAST) {
        u32 dwMacCr=Lan_GetRegDW(MAC_CR);
        dwMacCr&=(~MAC_CR_PRMS_);
        dwMacCr|=MAC_CR_MCPAS_;
        dwMacCr&=(~MAC_CR_HPFILT_);
        Lan_SetRegDW(MAC_CR,dwMacCr);
        smsc9420_pci_push(pd);
        SMSC_TRACE(DBG_MCAST, "Receive all Multicast Enabled");
        return;
    }

    if(rxHashLo||rxHashHi) {
        u32 dwMacCr=Lan_GetRegDW(MAC_CR);
        dwMacCr&=(~MAC_CR_PRMS_);
        dwMacCr&=(~MAC_CR_MCPAS_);
        dwMacCr|=MAC_CR_HPFILT_;
        Lan_SetRegDW(HASHH,rxHashHi);
        Lan_SetRegDW(HASHL,rxHashLo);
        Lan_SetRegDW(MAC_CR,dwMacCr);
    }
    else
    {
        u32 dwMacCr=Lan_GetRegDW(MAC_CR);
        dwMacCr&=(~MAC_CR_PRMS_);
        dwMacCr&=(~MAC_CR_MCPAS_);
        dwMacCr&=(~MAC_CR_HPFILT_);
        Lan_SetRegDW(HASHH,0);
        Lan_SetRegDW(HASHL,0);
        Lan_SetRegDW(MAC_CR,dwMacCr);
        SMSC_TRACE(DBG_MCAST,"Receive own packets only.");
    }
}

static void set_phy_sleep(pPD pd, const ulong phy_down)
{

    if(phy_down)
    {
        pmt_rst_phy(pd);
        //disable AutoNegotiation
        Phy_SetRegW(pd, PHY_BCR,
                (Phy_GetRegW(pd, PHY_BCR) & ~(PHY_BCR_ANEG_EN_)));
        //disable autoMDIX
        Phy_SetRegW(pd, PHY_CSIR,
                (Phy_GetRegW(pd,PHY_CSIR) & ~(PHY_SCS_AUTOMDIX_EN_|
                                              PHY_SCS_AUTOMDIX_STATE_)
                 ));
        //disable all interrupt sources
        Phy_SetRegW(pd, PHY_INT_MASK,  0);
        //clear INT src 
        Phy_GetRegW(pd, PHY_INT_SRC);
        Phy_GetRegW(pd, PHY_INT_SRC);

        //put phy in sleep 
        Phy_SetRegW(pd, PHY_BCR, PHY_BCR_PWR_DWN_);

    }
    return;
}


static void set_mac_sleep(pPD pd, const ulong mac_down)
{
    if(mac_down)
    {
        rst_mac_dmac(pd);
    }
    return;
}
#ifdef CONFIG_PM


static ulong is_link_up(pPD pd)
{
    Phy_GetRegW(pd, PHY_BSR);
    return (PHY_BSR_LINK_STATUS_ & Phy_GetRegW(pd, PHY_BSR));
}

static void process_wake_and_hw_sleep(struct net_device *ndev)
{
    WAKEUP_FILTER sFilter;
    ulong wol_opts;
    ulong mac_down = 1;
    ulong phy_down = 1;
    pPD pd = (pPD)netdev_priv(ndev);
    //check if we are  expected to wake at all
    if(netif_running(ndev) && (wol_opts = GET_PD_WOLOPTS(pd)))
    {
        //process mac_wol
        //check if we have MAC WAKEs
        if(wol_opts & LAN9420_MAC_WAKE_MASK)
        {
            //check that the link is up 
            //if it's not up, and WAKE_PHY is not present
            //we should proceed as if the link is up
            if(is_link_up(pd) || (!(wol_opts & WAKE_PHY)))
            {
                SMSC_TRACE(DBG_PWR,"Setting WAKE_MAC events");

            //and wake on MAC event. I'd ignore this possibility.
            //we need mac 
            mac_down = 0;
                //we need phy 
            phy_down = 0;
            if((WAKE_BCAST | WAKE_MCAST | WAKE_ARP | WAKE_UCAST) &
                    wol_opts)
            {
                memset(&sFilter, 0, sizeof(sFilter));
                if(WAKE_BCAST &
                        wol_opts) 
                {
                    sFilter.dwFilter0Mask = 0x003F;
                    sFilter.dwCommand |= 
                        ((MAC_WAKE_FILTERCMD_ADDR_TYPE_MCAST_ |
                         MAC_WAKE_FILTERCMD_EN_) << 0);
                    sFilter.dwOffset |= 0x00;
                    sFilter.dwCRC_10 |= LAN9420_BCAST_WAKE_CRC;
                }
                if(WAKE_MCAST &
                        wol_opts)
                {
                    sFilter.dwFilter1Mask = 0x0007;
                    sFilter.dwCommand |= 
                        ((MAC_WAKE_FILTERCMD_ADDR_TYPE_MCAST_ |
                          MAC_WAKE_FILTERCMD_EN_) << 8);
                    sFilter.dwOffset |= 0x00;
                    sFilter.dwCRC_10 |= (LAN9420_MCAST_WAKE_CRC << 16);
                }
                if(WAKE_ARP &
                        wol_opts)
                {
                    sFilter.dwFilter2Mask = 0x003;
                    sFilter.dwCommand |=
                       ((MAC_WAKE_FILTERCMD_ADDR_TYPE_MCAST_ |
                        MAC_WAKE_FILTERCMD_EN_) << 16); 
                    sFilter.dwOffset = 0x00;
                    sFilter.dwCRC_32 |= LAN9420_ARP_WAKE_CRC;
                }
                if(WAKE_UCAST &
                        wol_opts)
                {
                    sFilter.dwFilter3Mask = 0x003F;
                    sFilter.dwCommand |=
                        (MAC_WAKE_FILTERCMD_EN_ << 24);
                    sFilter.dwOffset |= 0x00;
                    sFilter.dwCRC_32 |= (CalculateCrc16(&(ndev->dev_addr[0]), 6, 0) << 16);
                }
                Lan_SetRegDW(WUFF, sFilter.dwFilter0Mask);
                Lan_SetRegDW(WUFF, sFilter.dwFilter1Mask);
                Lan_SetRegDW(WUFF, sFilter.dwFilter2Mask);
                Lan_SetRegDW(WUFF, sFilter.dwFilter3Mask);
                Lan_SetRegDW(WUFF, sFilter.dwCommand);
                Lan_SetRegDW(WUFF, sFilter.dwOffset);
                Lan_SetRegDW(WUFF, sFilter.dwCRC_10);
                Lan_SetRegDW(WUFF, sFilter.dwCRC_32);

                Lan_SetRegDW(WUCSR, Lan_GetRegDW(WUCSR)|
                        WUCSR_WAKE_EN_);//Enable Wake-up frame detection
            }
            if(WAKE_MAGIC &
                    wol_opts)
            {
                Lan_SetRegDW(WUCSR, 
                        Lan_GetRegDW(WUCSR) |
                        WUCSR_MPEN_);
            }

            //Enable detection of wake-up frame or magic packet
            Lan_SetRegDW(PMT_CTRL, Lan_GetRegDW(PMT_CTRL)|
                    PMT_CTRL_WUPS_MAC_);        //CLEAR WUPS[1]
            Lan_SetRegDW(PMT_CTRL, Lan_GetRegDW(PMT_CTRL)|PMT_CTRL_WOL_EN_);
            Lan_DisableInterrupt(pd,INT_CTL_PMT_INT_EN_); //disable it before sleeping 
            SET_PD_NEEDSWAKE(pd); 
            //enable recevier   
            Lan_SetRegDW(MAC_CR,Lan_GetRegDW(MAC_CR)|MAC_CR_RXEN_);         
            smsc9420_pci_push(pd);

            }
        }
#ifdef USE_WAKE_PHY_ON_PCI
        if(wol_opts & WAKE_PHY)
        {

            phy_down = 0;
            //read INT src twice, 
            //do not know why.
            //taken from NDIS driver. needs clarification
            Phy_GetRegW(pd, PHY_INT_SRC);
            Phy_GetRegW(pd, PHY_INT_SRC);
            //we should wake on PHY event
            if(is_link_up(pd))
            {

                //let's wake on link change?
                //is this a valid event?
                //disalbe EDPWRDOWN
                SMSC_TRACE(DBG_PWR,"Enabling LINK_CHANGE WAKE_PHY");
                Phy_SetRegW(pd, PHY_MCSR, 0);
                Phy_SetRegW(pd, PHY_INT_MASK, 
                        PHY_INT_MASK_ANEG_COMPL_ |
                        PHY_INT_MASK_REMFLT_ |
                        PHY_INT_MASK_LINK_DOWN_ );


            }
            else        //link is not up
            {
                //can't rst mac after we enter power sleep 
                //let's rst now

                set_mac_sleep(pd, mac_down);
                //prevent another attempt at function exit.
                mac_down = 0;
                SMSC_TRACE(DBG_PWR, "Enabling ENERGYON WAKE_PHY");
                //for ENERGYON to work
                //it is important to disable ANEG
                //and AUTOMDIX
                //the following sequence is taken from 
                //Woojung's 118 CE driver 

                //disable AutoNegotiation
                Phy_SetRegW(pd, PHY_BCR, 
                        (Phy_GetRegW(pd, PHY_BCR) & ~(PHY_BCR_ANEG_EN_)));
                //disable autoMDIX
                Phy_SetRegW(pd, PHY_CSIR, 
                        (Phy_GetRegW(pd,PHY_CSIR) & ~(PHY_SCS_AUTOMDIX_EN_|
                                                      PHY_SCS_AUTOMDIX_STATE_)
                        ));
                //clear INT src again
                Phy_GetRegW(pd, PHY_INT_SRC);
                //enable ENERGYON interrupt source
                Phy_SetRegW(pd, PHY_INT_MASK, (Phy_GetRegW(pd, PHY_INT_MASK) |
                            PHY_INT_MASK_ENERGYON_));
                //enable energy detect power-down mode
                Phy_SetRegW(pd, PHY_MCSR, PHY_MCSR_EDPWRDOWN_);
            }
                //enable energy-detect wake-up and clear wake-up status
                Lan_SetRegDW(PMT_CTRL, (Lan_GetRegDW(PMT_CTRL)  |
                                PMT_CTRL_WUPS_PHY_));   //clear WUPS[0]
                Lan_SetRegDW(PMT_CTRL, (Lan_GetRegDW(PMT_CTRL)  |
                                PMT_CTRL_ED_EN_ ));

               SET_PD_NEEDSWAKE(pd); 



        }
#endif
    }
    else
    {
    //there is no wolopts to support  or interface is not running

    }
    set_mac_sleep(pd, mac_down);
    set_phy_sleep(pd, phy_down);
    return;

}





static void ResetWakeupEvents(struct net_device *dev)
{

    pPD pd=(pPD)netdev_priv(dev);
    
    SMSC_ASSERT(pd!=NULL);
    SMSC_TRACE(DBG_PWR,"In ResetWakeupEvents. \n");

    if(!(GET_PD_WOLOPTS(pd)& ( WAKE_UCAST | WAKE_BCAST | 
                               WAKE_MCAST | WAKE_ARP | 
                               WAKE_MAGIC |
#ifdef USE_WAKE_PHY_ON_PCI
                               WAKE_PHY   |
#endif
                               0)))
    {
        return;
    }
    
    
    if(GET_PD_WOLOPTS(pd)& WAKE_PHY){
#ifdef USE_WAKE_PHY_ON_PCI
        //Disable the energy detect power-down mode
        Phy_SetRegW(pd,PHY_MCSR,Phy_GetRegW(pd,PHY_MCSR) & ~PHY_MCSR_EDPWRDOWN_);
        //Disable ENERGYON interrupt source
        Phy_SetRegW(pd,PHY_INT_MASK,Phy_GetRegW(pd,PHY_INT_MASK) & ~PHY_INT_MASK_ENERGYON_);
        //Disable energy-detect wake-up
        Lan_SetRegDW(PMT_CTRL,Lan_GetRegDW(PMT_CTRL) & ~PMT_CTRL_ED_EN_);   
#endif
    }else{

        if(GET_PD_WOLOPTS(pd)& (WAKE_BCAST | WAKE_MCAST | WAKE_ARP | WAKE_UCAST)){
            Lan_SetRegDW(WUCSR, Lan_GetRegDW(WUCSR)&~WUCSR_WAKE_EN_);//Disable Wake-up frame detection
        }
        if(GET_PD_WOLOPTS(pd)& WAKE_MAGIC){//Set Magic packet detection
            Lan_SetRegDW(WUCSR, Lan_GetRegDW(WUCSR)&~WUCSR_MPEN_);//Disable magic frame detection
        }
        //Disable wake-up frame interrupt
        Lan_SetRegDW(PMT_CTRL, Lan_GetRegDW(PMT_CTRL)&~PMT_CTRL_WOL_EN_);       
    }
    Lan_DisableInterrupt(pd,INT_CTL_PMT_INT_EN_);
    return;
}


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
static int Smsc9420_suspend (struct pci_dev *pdev, u32 state)
#else
static int Smsc9420_suspend (struct pci_dev *pdev, pm_message_t state)
#endif
{
    struct net_device *ndev = pci_get_drvdata(pdev);
    pPD pd=(pPD)netdev_priv(ndev);     //netdev_priv(dev);

//disable all our interrupts
    SMSC_TRACE(DBG_PWR, "%s(): for pci dev: %x\n",
            __FUNCTION__, pdev->device);
    Lan_DisableIRQ(pd);
    CLR_PD_NEEDSWAKE(pd);
    if(netif_running(ndev))
    {
        SMSC_TRACE(DBG_PWR, "suspending running iface on pci dev: %x\n",
                pdev->device);
        smsc9420_down(ndev);
        netif_device_detach(ndev);
        process_wake_and_hw_sleep(ndev);
    }
    else
    {
        SMSC_TRACE(DBG_PWR, "suspending inactive iface on pci dev: %x\n",
                pdev->device);
       // set_phy_sleep(pd, 1);
    }
        
    if(pci_save_state(pdev
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
               ,pd->PCI_conf_copy
#endif
                ))
    {
        SMSC_WARNING("pci_save_state() failed in %s():%d\n",
                __FUNCTION__, __LINE__);
    }


    if(pci_enable_wake(pdev, 
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
            state,
#else
            pci_choose_state(pdev,state),
#endif
            IS_PD_NEEDSWAKE_SET(pd)))
    {
        SMSC_WARNING("pci_enable_wake() failed in %s()", 
                __FUNCTION__);
    }
    pci_disable_device (pdev);
    if(pci_set_power_state(pdev, 
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
                state
#else
                pci_choose_state(pdev,state)
#endif
                ))
    {
        SMSC_WARNING("pci_set_power_state() failed in %s()",
                __FUNCTION__);

    }
    SMSC_TRACE(DBG_PWR, "%s():done for pci dev: %x\n", 
            __FUNCTION__, pdev->device);

    return 0;
}


static int Smsc9420_resume(struct pci_dev *pdev)
{
    struct net_device *ndev = pci_get_drvdata(pdev);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
    pPD pd=(pPD)netdev_priv(ndev);   //dev_priv(dev);
#endif
    int err;


    SMSC_TRACE(DBG_INIT,"--> %s()", __FUNCTION__);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
    pci_set_power_state(pdev, 0);
    pci_restore_state(pdev, pd->PCI_conf_copy);
#else
    pci_set_power_state(pdev, PCI_D0);
    pci_restore_state(pdev);
#endif
    err = pci_enable_device (pdev);
    if (err)
         return err;
    pci_set_master(pdev);

    err = pci_enable_wake(pdev,0,0);
    if (err) 
    {
        SMSC_WARNING("pci_enable_wake() failed in Smsc9420_resume:: err=%d", err);
    }
    ResetWakeupEvents(ndev);
    if(!netif_running(ndev))
    {
        return 0;
    }
    Smsc9420_open(ndev);
    netif_device_attach(ndev);
    return 0;
}

static u16 CalculateCrc16(const u8 * bpData,const ulong dwLen, const ulong fBitReverse)
{
    const u16 wCrc16Poly = 0x8005U;    // s/b international standard for CRC-16
                                // x^16 + x^15 + x^2 + 1
    u16 i, j, bit;
    u16 wCrc = 0xFFFFU;
    u16 wMsb;
    u8 bCurrentByte;
    u16 wNumOfBits = 16U;
    u16 wCrcOut=0;

    wNumOfBits = wNumOfBits; // to avoid lint warning

    for (i=0; i<(u16)dwLen; i++)
    {
        bCurrentByte = *bpData++;

        for (bit=(u16)0U; bit<(u16)8U; bit++)
        {
            wMsb = wCrc >> 15;
            wCrc <<= 1;

            if (wMsb ^ (u16)(bCurrentByte & 1))
            {
                wCrc ^= wCrc16Poly;
                wCrc |= (u16)0x0001U;
            }
            bCurrentByte >>= 1;
        }
    }
    //bit reverse if needed
    // so far we do not need this for 117
    // but the standard CRC-16 seems to require this.
    if (fBitReverse)
    {
        j = 1;
        for (i=(u16)(1<<(wNumOfBits-(u16)1U)); i; i = i>>1) {
            if (wCrc & i)
            {
                wCrcOut |= j;
            }
            j <<= 1;
        }
        wCrc = wCrcOut;
    }

    return wCrc;
}


#endif /* CONFIG_PM */

static int __init Smsc9420_init_module(void)
{
    
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13))
    return pci_module_init(&Smsc9420_driver);
#else
    return pci_register_driver(&Smsc9420_driver);
#endif
    /* return 0; */
}

static void __exit Smsc9420_exit_module(void)
{
    pci_unregister_driver(&Smsc9420_driver);
} 

module_init(Smsc9420_init_module);
module_exit(Smsc9420_exit_module);

